/*
 * NNTPFWD.C
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <assert.h>

typedef struct Collect {
    char	*c_MsgId;
    char	*c_Buf;
    int		c_Size;
    int		c_MaxSize;
    int		c_HasNewsGroup;
    int		c_HasListMail;
} Collect;

static FILE *NNTPConnect(void);
static void fatal(const char *ctl, ...);
static void DoPush(FILE *fp);
static void DoPull(FILE *fp);
static void DoRunQ(void);
static void nntpsend(FILE *fp, const  char *ctl, ...);
static int nntprecv(FILE *fp, char *buf, int size);
static char *whitestrip(char *ptr, int len);

static void init_collect(Collect *cc);
static void start_collect(Collect *cc);
static void reset_collect(Collect *cc);
static void free_collect(Collect *cc);
static void add_collect(Collect *cc, char *buf, int len);
static void end_collect_dumptonews(Collect *cc, FILE *fp);
static void end_collect_dumptomail(Collect *cc, int artno);

char *QueueDir;
char *NNTPHost;
char *NNTPGroup;
char *ListMail;

int
main(int ac, char **av)
{
    int i;
    enum { PUSH, PULL, RUNQ } mode = -1;
    FILE *fp;

    if (ac == 1) {
	fatal("%s [-q queuedir] [-h group@host] [-e mail@list] [-push/-pull/-runq]", av[0]);
    }

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];

	if (*ptr != '-') {
	    fatal("Expected option: %s", ptr);
	    /* not reached */
	}
	ptr += 2;
	switch(ptr[-1]) {
	case 'q':
	    QueueDir = av[++i];
	    break;
	case 'h':
	    NNTPGroup = av[++i];
	    if (NNTPGroup && (NNTPHost = strchr(NNTPGroup, '@')) != NULL)
		*NNTPHost++ = 0;
	    else
		fatal("Expected group@host for -h: %s", NNTPGroup);
	    break;
	case 'e':
	    ListMail = av[++i];
	    break;
	case 'p':	/* pull, push */
	    if (strcmp(ptr - 1, "pull") == 0)
		mode = PULL;
	    else if (strcmp(ptr - 1, "push") == 0)
		mode = PUSH;
	    else
		fatal("Unknown option: %s", ptr - 2);
	    break;
	case 'r':	/* runq */
	    if (strcmp(ptr - 1, "runq") == 0)
		mode = RUNQ;
	    else
		fatal("Unknown option: %s", ptr - 2);
	    break;
	default:
	    fatal("Unknown option: %s", ptr - 2);
	    /* not reached */
	}
    }
    if (i > ac)
	fatal("Expected argument to last option");
    if (QueueDir == NULL)
	fatal("Queue directory must be specified with -q in all modes");
    if (mode != RUNQ && NNTPGroup == NULL)
	fatal("Nntp group@host must be specified with -h when pushing or pulling");
    if (mode == PULL && ListMail == NULL)
	fatal("Listmail address must be specified with -h when pulling");

    switch(mode) {
    case PUSH:	/* push mail to nntp server */
	fp = NNTPConnect();
	DoPush(fp);
	break;
    case PULL:	/* pull articles from nntp server */
	fp = NNTPConnect();
	DoPull(fp);
	break;
    case RUNQ:	/* push queued mail to nntp server */
	DoRunQ();
	break;
    default:
	fatal("Must specify -push, -pull, or -runq");
    }
    return(0);
}

void
DoPush(FILE *fp)
{
    int n;
    int rc;
    int nart;
    int fart;
    int lart;
    int state = 0;
    char *base;
    char buf[1024];
    Collect cc;

    init_collect(&cc);

    nntpsend(fp, "group %s", NNTPGroup);
    nntprecv(fp, buf, sizeof(buf));
    n = sscanf(buf, "%d %d %d %d", &rc, &nart, &fart, &lart);
    if (n != 4 || rc != 211)
	fatal("DoPush: nntprecv: bad response to group command: %s", buf);
    while (fgets(buf, sizeof(buf), stdin) != NULL) {
	if (strncmp(buf, "From ", 5) == 0) {
	    if (state)
		end_collect_dumptonews(&cc, fp);
	    start_collect(&cc);
	    state = 1;
	    add_collect(&cc, "EnvelopeFrom: ", 14);
	    add_collect(&cc, buf + 5, strlen(buf + 5));
	} else if (state) {
	    if (strncmp(buf, ">From ", 6) == 0)
		base = buf + 1;
	    else
		base = buf;
	    n = strlen(base);
	    add_collect(&cc, base, n);
	}
    }
    if (state)
	end_collect_dumptonews(&cc, fp);
    free_collect(&cc);
}

void
DoPull(FILE *fp)
{
    int nart;
    int fart;
    int lart;
    int rc;
    int i;
    int n;
    int lnl = 0;
    FILE *fi;
    char *base;
    char buf[1024];
    Collect cc;

    init_collect(&cc);

    nntpsend(fp, "group %s", NNTPGroup);
    nntprecv(fp, buf, sizeof(buf));
    n = sscanf(buf, "%d %d %d %d", &rc, &nart, &fart, &lart);
    if (n != 4 || rc != 211)
	fatal("DoPull: nntprecv: bad response to group command: %s", buf);

    /*
     * Figure out which article to start with using tracking queue file.
     */
    snprintf(buf, sizeof(buf), "%s/N.%s", QueueDir, NNTPGroup);
    if ((fi = fopen(buf, "r+")) != NULL) {
	if (fscanf(fi, "%d", &nart) != 1)
	    nart = lart + 1;
    } else if ((fi = fopen(buf, "w")) != NULL) {
	nart = lart + 1;
    } else {
	fatal("Unable to create %s", buf);
    }
    printf("group %s %d-%d starting with %d\n", NNTPGroup, fart, lart, nart);
    fart = nart;
    for (i = fart; i <= lart; ++i) {
	nntpsend(fp, "article %d", i);
	rc = nntprecv(fp, buf, sizeof(buf));
	n = sscanf(buf, "%d %d", &rc, &nart);
	if (n == 2 && rc == 220 && nart == i) {
	    printf(" %s: article %d\n", NNTPGroup, i);
	    start_collect(&cc);
	    while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (lnl && strcmp(buf, ".\r\n") == 0) {
		    end_collect_dumptomail(&cc, i);
		    break;
		}
		if (buf[0] == '.' && buf[1] == '.')	/* dot escaped? */
		    base = buf + 1;
		else
		    base = buf;
		n = strlen(base);
		add_collect(&cc, base, n);
		if (n && base[n-1] == '\n')
		    lnl = 1;
		else
		    lnl = 0;
	    }
	    free_collect(&cc);
	}
    }
    fseek(fi, 0L, 0);
    fprintf(fi, "%-9d\n", i);
    fflush(fi);
    if (ferror(fi))
	fprintf(stderr, "Unable to update N.<group> queue file!\n");
    fclose(fi);
}

void
DoRunQ(void)
{
}

static
FILE *
NNTPConnect(void)
{
    int fd;
    int rc;
    FILE *fp;
    struct hostent *he;
    struct sockaddr_in sock;
    char buf[256];

    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	fatal("socket: %s", strerror(errno));
    if ((he = gethostbyname(NNTPHost)) == NULL)
	fatal("could not resolve host: %s", NNTPHost);
    sock.sin_family = he->h_addrtype;
    bcopy(he->h_addr, &sock.sin_addr, he->h_length);
    sock.sin_port = 0;
    if (bind(fd, (void *)&sock, sizeof(sock)) < 0)
	fatal("bind: %s", strerror(errno));
    sock.sin_port = htons(119);
    if (connect(fd, (void *)&sock, sizeof(sock)) < 0)
	fatal("connect: %s", strerror(errno));
    fp = fdopen(fd, "r+");
    rc = nntprecv(fp, buf, sizeof(buf));
    if (rc != 200) 
	fatal("Expected code 200, got: %s", buf);
    return(fp);
}

/*
 * COLLECTION CODE
 */

static void
init_collect(Collect *cc)
{
    bzero(cc, sizeof(Collect));
}

static void
free_collect(Collect *cc)
{
    reset_collect(cc);
    if (cc->c_Buf) {
	free(cc->c_Buf);
	cc->c_Buf = NULL;
    }
    cc->c_MaxSize = 0;
}

static void
start_collect(Collect *cc)
{
}

static void
reset_collect(Collect *cc)
{
    if (cc->c_MsgId) {
	free(cc->c_MsgId);
	cc->c_MsgId = NULL;
    }
    cc->c_Size = 0;
    cc->c_HasNewsGroup = 0;
    cc->c_HasListMail = 0;
}

static void
add_collect(Collect *cc, char *buf, int len)
{
    char *base;

    while (cc->c_Size + len >= cc->c_MaxSize) {
	cc->c_MaxSize += (cc->c_MaxSize / 2 + len + 4095) & ~4095;
	if (cc->c_Buf)
	    cc->c_Buf = realloc(cc->c_Buf, cc->c_MaxSize);
	else
	    cc->c_Buf = malloc(cc->c_MaxSize);
	assert(cc->c_Buf != NULL);
    }
    base = cc->c_Buf + cc->c_Size;
    bcopy(buf, base, len);

    /*
     * Strip CR in CRLF sequence
     */
    if (len >= 2 && base[len-2] == '\r' && base[len-1] == '\n') {
	base[len-2] = '\n';
	--len;
    }
    cc->c_Size += len;
    cc->c_Buf[cc->c_Size] = 0;

    /*
     * Remove To: and Cc: headers.
     */
    if (strncasecmp(base, "To:", 3) == 0) {
	cc->c_Size -= len;
	cc->c_Buf[cc->c_Size] = 0;
	return;
    }
    if (strncasecmp(base, "Cc:", 3) == 0) {
	cc->c_Size -= len;
	cc->c_Buf[cc->c_Size] = 0;
	return;
    }

    /*
     * Extract Message-Id: and check to see if the message has already
     * been posted to a newsgroup or mailing list (to avoid loops).
     */
    if (strncasecmp(base, "Message-Id:", 11) == 0) {
	if (cc->c_MsgId)
	    free(cc->c_MsgId);
	cc->c_MsgId = whitestrip(base + 11, len - 11);
    }
    if (strncasecmp(base, "Newsgroups:", 11) == 0) {
	cc->c_HasNewsGroup = 1;
	/*if (strcasestr(base, NNTPGroup) != NULL)*/
    }
    if (strncasecmp(base, "BestServHost:", 13) == 0 ||
	strncasecmp(base, "List-Post:", 10) == 0
    ) {
	cc->c_HasListMail = 1;
    }
}

static void
end_collect_dumptonews(Collect *cc, FILE *fp)
{
    int rc;
    char buf[1024];
    char *base;
    char *next;

    if (cc->c_Size && cc->c_HasNewsGroup == 0) {
	/*
	 * Check the message-id
	 */
	if (cc->c_MsgId) {
	    nntpsend(fp, "stat %s", cc->c_MsgId);
	    rc = nntprecv(fp, buf, sizeof(buf));
	    if (rc == 223) {
		printf("Duplicate %s\n", cc->c_MsgId);
		return;
	    }
	}
	/*
	 * Post it
	 */
	nntpsend(fp, "post");
	rc = nntprecv(fp, buf, sizeof(buf));
	if (rc == 340) {
	    fprintf(fp, "Newsgroups: %s\r\n", NNTPGroup);
	    base = cc->c_Buf;
	    while ((next = strchr(base, '\n')) != NULL) {
		if (base[0] == '.')
		    fprintf(fp, ".");
		fprintf(fp, "%*.*s\r\n", next - base, next - base, base);
		base = next + 1;
	    }
	    if (*base) {
		if (base[0] == '.')
		    fprintf(fp, ".");
		fprintf(fp, "%s\r\n", base);
	    }
	    fprintf(fp, ".\r\n");
	    fflush(fp);
	    rc = nntprecv(fp, buf, sizeof(buf));
	    printf("Posting RC %d %s\n", rc, buf);
	} else {
	    printf("Queueing %s\n", cc->c_MsgId);
	    /* queue it */
	}
    } else if (cc->c_MsgId) {
	printf("Will not repost, came from newsgroup: %s\n", cc->c_MsgId);
    }
    reset_collect(cc);
}

static void
end_collect_dumptomail(Collect *cc, int artno)
{
    if (cc->c_Size && cc->c_HasListMail == 0) {
	FILE *fp = popen("/usr/sbin/sendmail -t", "r+");
	fprintf(fp, "To: %s\n", ListMail);
	fwrite(cc->c_Buf, cc->c_Size, 1, fp);
	fflush(fp);
	fclose(fp);
    }
    reset_collect(cc);
}

/*
 * MISC SUPPORT
 */
static void
nntpsend(FILE *fp, const char *ctl, ...)
{
    va_list va;

    va_start(va, ctl);
    vfprintf(fp, ctl, va);
    va_end(va);
    fprintf(fp, "\r\n");
    fflush(fp);
}

static int
nntprecv(FILE *fp, char *buf, int size)
{
    if (fgets(buf, size, fp) == NULL)
	fatal("Unexpected EOF while getting nntp response");
    return(strtol(buf, NULL, 10));
}

static
void
fatal(const char *ctl, ...)
{
    va_list va;

    va_start(va, ctl);
    vfprintf(stderr, ctl, va);
    fprintf(stderr, "\n");
    va_end(va);
    exit(1);
}

static
char *
whitestrip(char *ptr, int len)
{
    char *res;

    while (len && (*ptr == ' ' || *ptr == '\t')) {
	--len;
	++ptr;
    }
    while (len && (ptr[len-1] == ' ' || ptr[len-1] == '\t' || 
	ptr[len-1] == '\r' || ptr[len-1] == '\n')
    ) {
	--len;
    }
    res = malloc(len + 1);
    assert(res);
    bcopy(ptr, res, len);
    res[len] = 0;
    return(res);
}

