
/*
 * AUTOQUEUE.C	-moves mail out of /var/spool/mqueue
 *
 * Mail is moved from /var/spool/mqueue/xxx to ../Qn/xxx, where 'n'
 * is the same as the last digit in xxx, thus spliting mail into 
 * 10 separate directories.
 *
 * Nominal usage is to call this program with options such that mail
 * that cannot be delivered within an hour is split off onto the slower
 * queues.
 *
 * SPECIAL CONSIDERATION:
 *	
 *	we only move non-zero length qf files that we can lock, and their
 *	associated data files.
 *
 *	0 length qf files may not have been locked by their creator yet
 *
 *	we do not move control files to a destination if there is a filename
 *	conflict.
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/dir.h>


void dorename(const char *qf, const char *prefix, int n);
int timeok(int fd, int dt);
int OkToMove(int fd);

time_t T;
int VerboseOpt;
int MaxMove = -1;
char *QBase = "../Q";

int
main(int ac, char **av)
{
    DIR *dir;
    struct direct *den;
    char *sdir = "/var/spool/mqueue";
    int i;
    int dt = 0;
    int redo = 0;
    int myqueue;

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];

	if (*ptr != '-') {
	    sdir = ptr;
	    continue;
	}
	ptr += 2;
	switch(ptr[-1]) {
	case 'n':
	    MaxMove = strtol(ptr, NULL, 0);
	    break;
	case 'd':
	    QBase = (*ptr) ? ptr : av[++i];
	    break;
	case 'r':
	    redo = strtol(ptr, &ptr, 0);

	    switch(*ptr) {
	    case 'm':
		redo *= 60;
		break;
	    case 'h':
		redo *= 60 * 60;
		break;
	    case 'd':
		redo *= 24 * 60 * 60;
		break;
	    }
	    break;
	case 'w':
	    dt = strtol(ptr, &ptr, 0);

	    switch(*ptr) {
	    case 'm':
		dt *= 60;
		break;
	    case 'h':
		dt *= 60 * 60;
		break;
	    case 'd':
		dt *= 24 * 60 * 60;
		break;
	    }
	    break;
	case 'v':
	    VerboseOpt = 1;
	    break;
	default:
	    fprintf(stderr, "unknown option: %s\n", ptr - 2);
	    exit(1);
	    break;
	}
    }

    if (redo) {
	for (;;) {
	    int pid;

	    if ((pid = fork()) == 0)
		break;
	    if (pid > 0)
		waitpid(pid, NULL, 0);
	    sleep(redo);
	}
	/* not reached */
    }

    T = time(NULL);

    myqueue = strtol(sdir + strlen(sdir) - 1, NULL, 0);

    if (chdir(sdir) != 0) {
	fprintf(stderr, "Unable to chdir to %s: %s\n", sdir, strerror(errno));
	exit(1);
    }

    if ((dir = opendir(".")) == NULL) {
	fprintf(stderr, "Can't read directory: %s\n", sdir);
	exit(1);
    }
    while ((den = readdir(dir)) != NULL) {
	if (strncmp(den->d_name, "qf", 2) == 0) {
	    int fd = open(den->d_name, O_RDWR);

	    if (fd >= 0) {
		struct stat st;

		if (fstat(fd, &st) == 0 && st.st_size > 0 && timeok(fd, dt)) {
		    if (flock(fd, LOCK_EX|LOCK_NB) >= 0) {
			int n;

			if ((n = OkToMove(fd)) != myqueue) {
				/*
			    int n = strtol(
				den->d_name + strlen(den->d_name) - 1, 
				NULL, 
				10
			    );
				*/
			    if (renameok(den->d_name, n)) {
				dorename(den->d_name, "df", n);
				dorename(den->d_name, "xf", n);
				dorename(den->d_name, "qf", n);
				if (VerboseOpt)
				    printf("Q%d <- %s\n", n, den->d_name);
			    } else {
				if (VerboseOpt)
				    printf("D%d <- %s\n", n, den->d_name);
			    }
			} else {
			    if (VerboseOpt)
				printf("n0 <- %s\n", den->d_name);
			}
		    } else if (VerboseOpt) {
			if (VerboseOpt)
			    printf("-- <- %s\n", den->d_name);
		    }
		} else {
		    if (VerboseOpt)
			printf("-- <- %s\n", den->d_name);
		}
		close(fd);
	    }
	    if (MaxMove > 0 && --MaxMove == 0)
		break;
	}
    }
}

int
renameok(const char *qf, int n)
{
    char path[1024];
    struct stat st;

    sprintf(path, "%s%d/%s", QBase, n, qf);
    if (stat(path, &st) == 0)
	return(0);
    return(1);
}

void
dorename(const char *qf, const char *prefix, int n)
{
    char path1[1024];
    char path2[1024];

    sprintf(path1, "%s%s", prefix, qf + 2);
    sprintf(path2, "%s%d/%s%s", QBase, n, prefix, qf + 2);
    rename(path1, path2);
}

int
timeok(int fd, int dt)
{
    char buf[4096];
    char *ptr;
    int n;
    int r = 0;

    if (dt == 0)
	return(1);

    if ((n = read(fd, buf, sizeof(buf) - 1)) < 0)
	return(0);
    buf[n] = 0;
    ptr = buf;
    for (;;) {
	if (*ptr == 'T') {
	    long xt = (long)T - (long)strtoul(ptr + 1, NULL, 10);

	    if (VerboseOpt)
		printf("(%d/%d) ", xt, dt);

	    if (xt >= dt || xt < -dt)
		r = 1;
	    break;
	}
	if ((ptr = strchr(ptr, '\n')) == NULL)
	    break;
	++ptr;
    }
    return(r);
}

/*
 * Return 'N' field value
 */

int
OkToMove(int fd)
{
    char buf[4096];
    int n;
    int r = -1;

    if ((n = read(fd, buf, sizeof(buf) - 1)) > 0) {
	char *p;

	buf[n] = 0;
	if ((p = strstr(buf, "\nN")) != NULL) {
	    r = strtol(p + 2, NULL, 0);
	}
    }
    if (r > 9)
	r = 9;
    return(r);
}

