
/*
 * EXEC.C	- SUBSYSTEM SUPPORT ROUTINES
 */

#include <time.h>

#include "defs.h"

Prototype void HandlePolicy(Msg *msg, const char *polFlags, const char *modFlags, void (*func)(Msg *msg, int mod, const char *polFlags, const char *modFlags), int mod);
Prototype void DoRunDigest(MailList *list, Msg *msg);

void DoSubscribe(Msg *, const char *, int, char, char, char, int);
void DoGetConfig(Msg *msg);
const char *DoSetConfig(const char *base, Msg *msg, char *arg, int scanOnly);
void DoGetInfo(Msg *msg);
const char *DoSetInfo(const char *base, Msg *msg, char *arg, int scanOnly);
void DoHelp(Msg *msg, int mod);
void DoInfo(Msg *msg, int mod);
void SaveDigest(Msg *msg, int msgFromMod);
void DoSetDigestVol(Msg *msg, int v);
void DoSetDigestNo(Msg *msg, int v);
void DoGetList(Msg *msg, const char *cmd, char c);
void AuthSubscribe(Msg *msg, const char *arg, int alen, char digestFlag, char aliasFlag);
void CopyOptionalHeaders(Msg *msg, int msgFromMod, int fd, FILE *fo);
int CheckFromNNTP(Msg *msg);

void 
HandlePolicy(Msg *msg, const char *polFlags, const char *modFlags, void (*func)(Msg *msg, int mod, const char *polFlags, const char *modFlags), int mod)
{
    int msgFromMod = FROM_UNKNOWN;

    /*
     * Pre-authenticate message.  
     *
     *	FROM_UNKNOWN		not from moderator
     *	FROM_MODERATOR		apparently from moderator
     *	FROM_MODERATOR_AUTH	authenticated from moderator (has password)
     */

    if (msg)
	msgFromMod = MsgFromModerator(msg, NULL);

    /*
     * Determine if sender's domain ends in local domain, if local
     * domain policy is in effect.
     */
    if (strchr(polFlags, 'L') != NULL && msg->ms_State == MSSTATE_ACTIVE) {
    	const char *local = GetListVar(msg->ms_List, "localDomain", 0, NULL);
	int localLen = strlen(local);

	if (local != NULL) {
	    int len;
	    const char *email;

	    email = findHeader(msg, "From:", 0, &len, "");
	    email = parseAddress(email, len, &len);
	    email = parseEMailFromAddress(email, len, &len);
	    if (email == NULL || len == 0) {
		msg->ms_BounceFlag = BOUNCE_RETURN;
		msg->ms_DeleteFlag = 1;
		msg->ms_ErrorFlag = 1;

		msgprintf(msg, "Message failed, I was not able to parse the email address from\n");
		msgprintf(msg, "The From: header.\n");
		return;
	    }
	    if (len > 128) {
		msg->ms_BounceFlag = BOUNCE_RETURN;
		msg->ms_DeleteFlag = 1;
		msg->ms_ErrorFlag = 1;

		msgprintf(msg, "I'm sorry, your From: header is too long.  The header may not exceed\n");
		msgprintf(msg, "128 characters in length.\n");
		return;
	    }

	    /*
	     * Isolate the domain part of the email address
	     */
	    email = rindex( email, '@') + 1;
	    len = strlen( email);
	    if (len < localLen || strncasecmp(local, email + len - localLen, localLen)) {
		msg->ms_BounceFlag = BOUNCE_RETURN;
		msg->ms_DeleteFlag = 1;
		msg->ms_ErrorFlag = 1;

		msgprintf(msg, "I'm sorry, you are not a local user.\n");
		return;
	    }
	}
    }

    /*
     * S - subscriber only posting.
     *
     * Determine if subscribed to the list by parsing
     * From: field and looking it up
     */
    if (strchr(polFlags, 'S') != NULL && msg->ms_State == MSSTATE_ACTIVE) {
	int len;
	int ok = 0;
	const char *email;

	email = findHeader(msg, "From:", 0, &len, "");
	email = parseAddress(email, len, &len);
	email = parseEMailFromAddress(email, len, &len);
	if (email == NULL || len == 0) {
	    msg->ms_BounceFlag = BOUNCE_RETURN;
	    msg->ms_DeleteFlag = 1;
	    msg->ms_ErrorFlag = 1;

	    msgprintf(msg, "Message failed, I was not able to parse the email address from\n");
	    msgprintf(msg, "The From: header.\n");
	    return;
	}
	if (len > 128) {
	    msg->ms_BounceFlag = BOUNCE_RETURN;
	    msg->ms_DeleteFlag = 1;
	    msg->ms_ErrorFlag = 1;

	    msgprintf(msg, "I'm sorry, your From: header is too long.  The header may not exceed\n");
	    msgprintf(msg, "128 characters in length.\n");
	    return;
	}

	/*
	 * Allow the message if it is 'from' a moderator address
	 * or if it's an authentic moderator-sent message.
	 */

	if (msgFromMod == FROM_MODERATOR_AUTH)
	    ok = 1;

	if (msgFromMod == FROM_MODERATOR)
	    ok = 1;

	/*
	 * Compare against subscriber list if mod = 0 (not a moderator
	 * command request)
	 */

	if (ok == 0 && mod == 0) {
	    if (FindListAddress(msg->ms_List, email, len) == 0) {
		ok = 1;
	    }
	}

#ifdef NOTDEF
	/*
	 * Compare against moderator addresses
	 */
	{
	    int n;
	    const char *str;

	    for (
		n = 1; 
		(str = GetListVar(msg->ms_List, "moderator",n,NULL)) != NULL;
		++n
	    ) {
		if (strlen(str) == len && strncmp(str, email, len) == 0) {
		    ok = 1;
		    break;
		}
	    }
	}
#endif
	
	if (ok == 0) {
	    msg->ms_BounceFlag = BOUNCE_RETURN;
	    msg->ms_DeleteFlag = 1;
	    msg->ms_ErrorFlag = 1;

	    if (mod) {
		msgprintf(msg, "I'm sorry, the -command alias may only be used by the moderator\n");
		msgprintf(msg, "of the list\n");
	    } else if (DumpLines("msgNotSub", msg->ms_List, msg, NULL, -1) == 0) {
		msgprintf(msg, "I'm sorry, this list requires that you be a subscribed user to\n");
		msgprintf(msg, "execute this command or post to the list.\n");
	    }
	    return;
	}
    }

    /*
     * Snip flame wars
     */
    if (CheckFlameMail(msg) < 0) {
	msg->ms_BounceFlag = BOUNCE_RETURN;
	msg->ms_DeleteFlag = 1;
	msg->ms_ErrorFlag = 1;

	msgprintf(msg,
	    "I'm sorry, the Subject line of your posting matches a thread\n"
	    "that has been terminated, probably because it turned into a\n"
	    "flame war.  No further postings with this subject (made by\n"
	    "anyone, not just you) are allowed.\n"
	    "\n"
	    "Please note that continuing a thread on the public list which\n"
	    "has been terminated in this manner by changing the subject line\n"
	    "can result in your email address being banned.\n"
	    "\n"
	);
	return;
    }

    /*
     * Y - Allow NNTP postings through without verification.
     *
     * X - Allow the posting through if subscribed, otherwise authenticate
     * with the originator.
     *
     * Determine if subscribed to the list by parsing
     * From: field and looking it up
     */
    if (strchr(polFlags, 'Y') != NULL && msg->ms_State == MSSTATE_ACTIVE && CheckFromNNTP(msg)) {
	/* 
	 * (do nothing) - allow nntp-sourced messages through without authentication.
	 */
    } else if (strchr(polFlags, 'X') != NULL && msg->ms_State == MSSTATE_ACTIVE) {
	int len;
	int ok = 0;
	const char *email;

	email = findHeader(msg, "From:", 0, &len, "");
	email = parseAddress(email, len, &len);
	email = parseEMailFromAddress(email, len, &len);
	if (email == NULL || len == 0) {
	    msg->ms_BounceFlag = BOUNCE_RETURN;
	    msg->ms_DeleteFlag = 1;
	    msg->ms_ErrorFlag = 1;

	    msgprintf(msg, "Message failed, I was not able to parse the email address from\n");
	    msgprintf(msg, "The From: header.\n");
	    return;
	}
	if (len > 128) {
	    msg->ms_BounceFlag = BOUNCE_RETURN;
	    msg->ms_DeleteFlag = 1;
	    msg->ms_ErrorFlag = 1;

	    msgprintf(msg, "I'm sorry, your From: header is too long.  The header may not exceed\n");
	    msgprintf(msg, "128 characters in length.\n");
	    return;
	}

	/*
	 * Allow the message if it is 'from' a moderator address
	 * or if it's an authentic moderator-sent message.
	 */

	if (msgFromMod == FROM_MODERATOR_AUTH)
	    ok = 1;

	if (msgFromMod == FROM_MODERATOR)
	    ok = 1;

	/*
	 * Compare against subscriber list if mod = 0 (not a moderator
	 * command request)
	 */
	if (ok == 0 && mod == 0) {
	    if (FindListAddress(msg->ms_List, email, len) == 0) {
		ok = 1;
	    }
	}

	/*
	 * The X option currently does not apply to moderator requests and
	 * should only be used for the posting policy.
	 */
	if (ok == 0 && mod == 0) {
	    msg->ms_BounceFlag = BOUNCE_AUTHENTICATE;

	    if (mod || DumpLines("msgXVerify", msg->ms_List, msg, NULL, -1) == 0) {
		msgprintf(msg, "Hello, this list requires that all postings from non-subscribed\n");
		msgprintf(msg, "addresses authenticate their postings.  If you are subscribed\n");
		msgprintf(msg, "but you use a different From: address you can send an 'alias'\n");
		msgprintf(msg, "command to <listname>-request@listserver specifying your alternate\n");
		msgprintf(msg, "email addresses.  The list will not send dups to your aliases, only\n");
		msgprintf(msg, "to your primary subscription address.\n");
		msgprintf(msg, "\n");
		msgprintf(msg, "If you wish me to post your message, simply reply to this message\n");
		msgprintf(msg, "changing the REJECT keyword to ACCEPT and leaving the remainder of\n");
		msgprintf(msg, "the Subject: line intact.\n");
		msgprintf(msg, "\n");
		msgprintf(msg, "If you did not send this message (i.e. someone is spoofing you),\n");
		msgprintf(msg, "you may wish to notify the list manager at the list server site.\n");
		msgprintf(msg, "If you do not wish this command to be executed or this posting\n");
		msgprintf(msg, "to be made, you can either respond and leave the REJECT intact,\n");
		msgprintf(msg, "or you can simply not respond at all\n");
	    }

	    msg->ms_DeleteFlag = 0;
	    return;
	}
    }

    /*
     * 'A' flag - authentication
     *
     * Send authentication request to moderator/user, leave message in queue.
     *
     * If 'mod' is set (moderator command), authentication is sent to the 
     * moderator rather then the sender of the message.  However, if the From: 
     * field of the message matches any 'moderatorXX' fields, that is where
     * the authentication will be sent.
     *
     * If the email came from a moderator address, we authenticate as if 'mod'
     * was set.  We also do this if a PASSWORD has been passed in the Subject:
     */

    if (strchr(polFlags, 'A') == NULL && msg->ms_State == MSSTATE_ACTIVE)
	msg->ms_State = MSSTATE_AUTHENTICATED1;

    if (strchr(polFlags, 'A') != NULL && msg->ms_State == MSSTATE_ACTIVE) {
	if (msgFromMod == FROM_MODERATOR_AUTH) {
	    /*
	     * we got a valid password, no additional authentication is required
	     */
	    msg->ms_State = MSSTATE_AUTHENTICATED2;
	    /* XXX lseek/write */
	} else if (mod || msgFromMod == FROM_MODERATOR) {
	    /*
	     * A moderator command or a message from the moderator... send authentication
	     * to the moderator and skip 'forward to moderator' step in next section when
	     * we get the response.
	     */
	    msg->ms_BounceFlag = BOUNCE_AUTHMODERATOR;
	    msg->ms_IncludeCompleteMessage = 1;

	    /*
	     * set new message id so originator cannot spoof the moderator 
	     */
	    ChangeMsgId(msg, MSSTATE_AUTHENTICATED2);

	    if (mod || DumpLines("msgVerify", msg->ms_List, msg, NULL, -1) == 0) {
		if (mod)
		    msgprintf(msg, "The following MODERATOR LEVEL command has been issued.  Please\n");
		else
		    msgprintf(msg, "The following request has been issued.  Please\n");
		msgprintf(msg, "verify the command by replying to this message and changing\n");
		msgprintf(msg, "the REJECT keyword into an ACCEPT keyword on the subject line\n");
	    }
	    msg->ms_DeleteFlag = 0;
	    return;
	} else {
	    msg->ms_BounceFlag = BOUNCE_AUTHENTICATE;

	    if (mod || DumpLines("msgVerify", msg->ms_List, msg, NULL, -1) == 0) {
		msgprintf(msg, "This list requires that commands be authenticated.  We received\n");
		msgprintf(msg, "the following message from your address.  If you wish me to\n");
		msgprintf(msg, "execute the command or post the post, simply reply to this\n");
		msgprintf(msg, "message changing the REJECT keyword to ACCEPT and leaving the\n");
		msgprintf(msg, "remainder of the Subject: line intact.\n");
		msgprintf(msg, "\n");
		msgprintf(msg, "If you did not send this message (i.e. someone is spoofing you),\n");
		msgprintf(msg, "you may wish to notify the list manager at the list server site.\n");
		msgprintf(msg, "If you do not wish this command to be executed or this posting\n");
		msgprintf(msg, "to be made, you can either respond and leave the REJECT intact,\n");
		msgprintf(msg, "or you can simply not respond at all\n");
	    }

	    msg->ms_DeleteFlag = 0;
	    return;
	}
    }

    /*
     * The message has been authenticated by the originator.  If this
     * facility is moderated, the message must now be authenticated by
     * the moderator... unless it already has been with a password
     */

    if ((msgFromMod == FROM_MODERATOR_AUTH || strchr(modFlags, 'Y') == NULL)
	&& msg->ms_State == MSSTATE_AUTHENTICATED1
    ) {
	msg->ms_State = MSSTATE_AUTHENTICATED2;
    }

    if (
	strchr(modFlags, 'Y') != NULL && 
	msg->ms_State == MSSTATE_AUTHENTICATED1
    ) {
	msg->ms_BounceFlag = BOUNCE_AUTHMODERATOR;
	msg->ms_IncludeCompleteMessage = 1;

	/*
	 * set new message id so originator cannot spoof the moderator 
	 */
	ChangeMsgId(msg, MSSTATE_AUTHENTICATED2);

	msgprintf(msg, "The following command or posting has been sent to the list.\n");
	msgprintf(msg, "Please verify the command by replying to this message and changing\n");
	msgprintf(msg, "the REJECT keyword into an ACCEPT keyword on the subject line.\n");
	msgprintf(msg, "To reject the message, leave the REJECT keyword intact or do not\n");
	msgprintf(msg, "reply to this message at all.\n");
	msgprintf(msg, "\n");
	msgprintf(msg, "The entire message is included below.  If as the moderator you\n");
	msgprintf(msg, "wish to make editorial changes, reply with a REJECT or ignore\n");
	msgprintf(msg, "this message, make the appropriate editorial changes, and post\n");
	msgprintf(msg, "the result independantly.  Any changes you make to an ACCEPT\n");
	msgprintf(msg, "will be ignored\n");

	msg->ms_DeleteFlag = 0;
	return;
    }

    /*
     * Execute function
     */

    func(msg, mod, polFlags, modFlags);

    /*
     * The moderator wants a carbon copy.  If we are going to send
     * the email to the moderator anyway, don't ALSO CC it to him.
     */

    if (strchr(modFlags, 'C') != NULL && 
	(msgFromMod == FROM_UNKNOWN || msg->ms_BounceFlag == 0)
    ) {
	msg->ms_BounceFlag |= BOUNCE_CCMODERATOR;
    }
}

/************************************************************************
 *                      EXECUTE MESSAGE OPERATION                       *
 ************************************************************************
 *
 */

Prototype void PostExec(Msg *msg, int mod, const char *polFlags, const char *modFlags);
Prototype void RequestExec(Msg *msg, int mod, const char *polFlags, const char *modFlags);

/*
 * Post the message to the list
 */
char *IncHead[] = {
    "Message-Id:",
    "Newsgroups:",
    "NNTP-Posting-Host:",
    "Xref:",
    "X-Trace:",
    "References:"
};

#if 0
char *DigHead[] = {
    "NNTP-Posting-Host:",
    "Xref:",
    "X-Trace:",
    "References:"
};
#endif

void
PostExec(Msg *msg, int mod, const char *polFlags, const char *modFlags)
{
    int hasDigest = 0;
    int maxMsgSize = GetListVal(msg->ms_List, "maxMsgSize", 0, 100*1024);
    int msgFromMod = MsgFromModerator(msg, NULL);
    int digestOnly = (GetListVar(msg->ms_List, "digestOnly", 0, "N")[0] == 'Y');

    /*
     * Do not accept messages from mailer-daemon@ or postmaster@ sent to
     * the main list, these are probably bounces from spammers using our
     * main list address as their From: address.
     */
    if (msgFromMod == FROM_UNKNOWN) {
	if (checkHeader(msg, "From:", "mailer-daemon@") >= 0) {
	    msg->ms_ErrorMsg = "Will not post messages from mailer-daemon";
	    return;
	}
    }

    /*
     * Do not accept message if it is too large.  Ignore size limit
     * if message is likely from a moderator.
     */

    if (msgFromMod == FROM_UNKNOWN && 
	maxMsgSize > 0 && 
	msg->ms_Body.he_Bytes >= maxMsgSize
    ) {
	msg->ms_BounceFlag = BOUNCE_RETURN;
	msgprintf(msg, "Your message is too large, messages cannot exceed\n");
	msgprintf(msg, "%d bytes\n", maxMsgSize);
	return;
    }

    /*
     * step 1 - administrivia
     *
     * locate a 'subscribe', 'info', or 'unsubscribe' request
     * in the first four lines of the message, but only if the
     * message is less then 15 lines long.
     */

    if (msgFromMod == FROM_UNKNOWN && GetListVar(msg->ms_List, "administrivia", 0, "Y")[0] == 'Y') {
	int lineNo = 1;
	int trivia = 0;
	const char *ptr = msg->ms_Body.he_Base;

	while (ptr) {
	    while (*ptr == ' ' || *ptr == '\t')
		++ptr;
	    if (lineNo <= 4) {
		if (strncasecmp(ptr, "info\n", 5) == 0 ||
		    strncasecmp(ptr, "info ", 5) == 0 ||
		    strncasecmp(ptr, "info\t", 5) == 0 ||
		    strncasecmp(ptr, "subscribe\n", 10) == 0 ||
		    strncasecmp(ptr, "subscribe ", 10) == 0 ||
		    strncasecmp(ptr, "subscribe\t", 10) == 0 ||
		    strncasecmp(ptr, "unsubscribe\n", 12) == 0 ||
		    strncasecmp(ptr, "unsubscribe ", 12) == 0 ||
		    strncasecmp(ptr, "unsubscribe\t", 12) == 0
		) {
		    trivia = 1;
		}
	    }
	    if ((ptr = strchr(ptr, '\n')) != NULL)
		++ptr;
	    ++lineNo;
	}

	/*
	 * Forward to moderator
	 */
	if (lineNo <= 15 && trivia) {
	    msg->ms_BounceFlag = BOUNCE_FWD_MODERATOR;
	    msg->ms_ErrorMsg = "Caught Administrivia";
	    msgprintf(msg, "The included message appears to be administrivia\n");
	    msgprintf(msg, "that was sent to the main list.  The server has\n");
	    msgprintf(msg, "forwarded the message to the moderator rather then\n");
	    msgprintf(msg, "post it to the list.  The commands have not been\n");
	    msgprintf(msg, "executed\n");
	    return;
	}
    }

    /*
     * step 2 - mail to non-digest list
     */
    
    if (digestOnly == 0) {
	int pid;
	int n;
	int fd = copenSendmail(&pid, msg->ms_List, "errors", 1);
	const char *ptr;
	const char *tp;
	int len;
	FILE *fi;

	/*
	 * Original Received: lines
	 */
	n = 0;
	while ((ptr = findHeader(msg, "Received:", n, &len, NULL)) != NULL) {
	    cprintf(fd, "Received: %*.*s\n", len, len, ptr);
	    ++n;
	}

	/*
	 * Headers to copy from original message, including Message-Id:,
	 * Newsgroups:, etc.
	 */
	for (n = 0; n < arysize(IncHead); ++n) {
	    if ((ptr = findHeader(msg, IncHead[n], 0, &len, NULL)) != NULL)
		cprintf(fd, "%s %*.*s\n", IncHead[n], len, len, ptr);
	}

	/*
	 * User's From: and Subject:
	 */
	ptr = findHeader2(msg, msgFromMod, "From:", &len, "");
	cprintf(fd, "From: %*.*s\n", len, len, ptr);
	ptr = findHeader2(msg, msgFromMod, "Mail-Followup-To:", &len, NULL);
	if (ptr)
	    cprintf(fd, "Mail-Followup-To: %*.*s\n", len, len, ptr);
	ptr = findHeader2(msg, msgFromMod, "Subject:", &len, "");
	ptr = stripPassword(ptr, len, &len);
	if ((tp = GetListVar(msg->ms_List, "subjectPrefix", 0, NULL)) != NULL && stringInString(tp, ptr, len) < 0)
	    cprintf(fd, "Subject: %s %*.*s\n", tp, len, len, ptr);
	else
	    cprintf(fd, "Subject: %*.*s\n", len, len, ptr);

	ptr = findHeader2(msg, msgFromMod, "Date:", &len, "");
	cprintf(fd, "Date: %*.*s\n", len, len, ptr);

	/*
	 * Our BestServHost: lines
	 */
	for (n=0; (ptr=findHeader(msg, "BestServHost:", n, &len, NULL)); ++n) {
	    cprintf(fd, "BestServHost: %*.*s\n", len, len, ptr);
	}
	cprintf(fd, "BestServHost: %s\n", HostName);

	/*
	 * Add our list headers.
	 */
	cprintf(fd,
		("List-Post: <mailto:%s@%s>\n"
		 "List-Subscribe: <mailto:%s-request@%s?body=subscribe>\n"
		 "List-Unsubscribe: <mailto:%s-request@%s?body=unsubscribe>\n"
		 "List-Help: <mailto:%s-request@%s?body=help>\n"
		 "List-Owner: <mailto:owner-%s@%s>\n"
		    ),
		msg->ms_ListName, HostName,
		msg->ms_ListName, HostName,
		msg->ms_ListName, HostName,
		msg->ms_ListName, HostName,
		msg->ms_ListName, HostName
	    );
	
	CopyOptionalHeaders(msg, msgFromMod, fd, NULL);

	/*
	 * Our Sender, Errors-To, and To field (note: the To: line actually
	 * causes the message to loop back to us, but the BestServHost:
	 * header catches it and we throw it away).
	 */

	cprintf(fd, "Sender: %s-errors@%s\n", msg->ms_ListName, HostName);
	cprintf(fd, "Errors-To: %s-errors@%s\n", msg->ms_ListName, HostName);
	ptr = findHeader2(msg, msgFromMod, "Reply-To:", &len, NULL);
	if (ptr) {
	    cprintf(fd, "Reply-To: %*.*s\n", len, len, ptr);
	} else if ((tp = GetListVar(msg->ms_List, "replyTo", 0, NULL)) != NULL
	    && tp[0]
	) {
	    cprintf(fd, "Reply-To: %s\n", tp);
	}
	cprintf(fd, "To: %s@%s\n", msg->ms_ListName, HostName);

	/*
	 * Generate Bcc: to the users on the list.  Is there a practical
	 * limit or not ?
	 */

	cprintf(fd, "Bcc:");

	if ((fi = OpenListFile(msg->ms_ListName, "list", 0, LOCK_SH)) != NULL) {
	    char buf[256];
	    char *efull;
	    char *flags;
	    int n = 0;
	    int noDigestOnly = (GetListVar(msg->ms_List, "noDigestOnly", 0, "N")[0] == 'Y');

	    while (GetListFile(fi, buf, sizeof(buf), &flags, &efull) == 0) {
		if (noDigestOnly == 0 && strchr(flags, 'd'))
		    hasDigest = 1;
		if ((noDigestOnly || strchr(flags, 'd') == NULL) && strchr(flags, 'a') == NULL) {
		    int elen;
		    const char *email;
		    
		    email = parseEMailFromAddress(efull, strlen(efull), &elen);
		    if (email != NULL && elen > 0) {
			if (n == 0) {
			    cprintf(fd, " %*.*s", elen, elen, email);
			} else {
			    cprintf(fd, ",\n\t%*.*s", elen, elen, email);
			}
			++n;
		    }
		}
	    }
	    fclose(fi);
	}
	cprintf(fd, "\n");	/* end of BCC */

	/*
	 * blank line, then the message body
	 */
	cprintf(fd, "\n");
	DumpLines("headerSing", msg->ms_List, NULL, NULL, fd);
	ptr = stripPasswordFromBody(msg->ms_Body.he_Base, msg->ms_Body.he_Bytes, &len);
	cwrite(fd, ptr, len);
	DumpLines("footerSing", msg->ms_List, NULL, NULL, fd);

	cclose(fd, pid);
    }

    /*
     * step 3 - save to digest if anyone is on the digest
     */

    if (hasDigest || digestOnly) {
	SaveDigest(msg, msgFromMod);
    }
}

/*
 * Command request for list or global
 */

void
RequestExec(Msg *msg, int mod, const char *polFlags, const char *modFlags)
{
    const char *base;
    char buf[256];
    char digestFlag = 0;
    char aliasFlag = 0;
    int didSetConfig = 0;
    int didSetInfo = 0;
    int didSubscribe = 0;

    if (msg->ms_ExecScanOnly == 0) {
	msg->ms_BounceFlag |= BOUNCE_RETURN;
	msg->ms_ErrorMsg = "Command/Request Response";
    }

    if (msg->ms_List) {
	if (GetListVar(msg->ms_List, "digestOnly", 0, "N")[0] == 'Y')
	    digestFlag = 'd';
    }

    base = msg->ms_Body.he_Base;
    while (base && *base) {
	int l;

	for (l = 0; base[l] && base[l] != '\n'; ++l)
	    ;
	if (l < sizeof(buf)) {
	    char *cmd;
	    char *arg;

	    bcopy(base, buf, l);
	    buf[l] = 0;

	    if (buf[0] != ' ' && buf[0] != '\t' && (cmd = strtok(buf, " \t\n\r")) != NULL) {
		int alen;

		if ((arg = strtok(NULL, "\n\r")) != NULL) {
		    while (*arg == ' ' || *arg == '\t')
			++arg;
		    for (alen = strlen(arg) - 1; alen >= 0; --alen) {
			if (arg[alen] != ' ' && arg[alen] != '\t')
			    break;
		    }
		    arg[alen+1] = 0;
		} else {
		    arg = "";
		}
		alen = strlen(arg);

		if (msg->ms_ExecScanOnly == 0 && strcasecmp(cmd, "approved:") != 0)
		    msgprintf(msg, "Command: %s %*.*s\n", cmd, alen, alen, arg);

		/*
		 * NON-MODERATOR COMMANDS
		 */

		if (strcasecmp(cmd, "approved:") == 0) {
		    ; /* no operation, handled somewhere else */
		} else if (strcasecmp(cmd, "help") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_GENERAL;
		    else
			DoHelp(msg, mod);
		} else if (strcasecmp(cmd, "info") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_GENERAL;
		    else
			DoInfo(msg, mod);
		    didSubscribe = -1;
		} else if (strcasecmp(cmd, "subsingle") == 0 ||
			   strcasecmp(cmd, "subscribe") == 0) {
		    if (msg->ms_ExecScanOnly) {
			msg->ms_ExecScanFlags |= EXSF_SUBSCRIBE;
		    } else if (mod || 
			    strchr(polFlags, 'R') == NULL || 
			    msg->ms_State == MSSTATE_AUTHSUBSCRIBE ||
			    alen == 0 ||
			    MsgFromCompare(msg, arg, alen) == 0
		    ) {
			DoSubscribe(msg, arg, alen, digestFlag,aliasFlag,'+', mod);
		    } else {
			AuthSubscribe(msg, arg, alen, digestFlag,aliasFlag);
		    }
		    if (didSubscribe == 0)
			didSubscribe = 1;
		} else if (strcasecmp(cmd, "alias") == 0) {
		    aliasFlag = 'a';
		    if (msg->ms_ExecScanOnly) {
			msg->ms_ExecScanFlags |= EXSF_SUBSCRIBE;
		    } else if (mod ||
			    strchr(polFlags, 'R') == NULL ||
			    msg->ms_State == MSSTATE_AUTHSUBSCRIBE ||
			    alen == 0 || 
			    MsgFromCompare(msg, arg, alen) == 0
		    ) {
			DoSubscribe(msg, arg, alen, digestFlag, aliasFlag, '+', mod);
		    } else {
			AuthSubscribe(msg, arg, alen, digestFlag, aliasFlag);
		    }
		    if (didSubscribe == 0)
			didSubscribe = 1;
		    aliasFlag = 0;
		} else if (strcasecmp(cmd, "subdigest") == 0) {
		    char oldDigestFlag = digestFlag;

		    digestFlag = 'd';
		    if (msg->ms_ExecScanOnly) {
			msg->ms_ExecScanFlags |= EXSF_SUBSCRIBE;
		    } else if (mod ||
			    strchr(polFlags, 'R') == NULL ||
			    msg->ms_State == MSSTATE_AUTHSUBSCRIBE ||
			    alen == 0 ||
			    MsgFromCompare(msg, arg, alen) == 0
		    ) {
			DoSubscribe(msg, arg, alen, digestFlag, aliasFlag, '+', mod);
		    } else {
			AuthSubscribe(msg, arg, alen, digestFlag, aliasFlag);
		    }
		    if (didSubscribe == 0)
			didSubscribe = 1;
		    digestFlag = oldDigestFlag;
		} else if (strcasecmp(cmd, "unsubscribe") == 0) {
		    if (msg->ms_ExecScanOnly) {
			msg->ms_ExecScanFlags |= EXSF_UNSUBSCRIBE;
		    } else {
			DoSubscribe(msg, arg, alen, 0, 0, '-', mod);
		    }
		} else if (strcasecmp(cmd, "end") == 0) {
		    /*
		     * end of commands, skip to end of body
		     */
		    l = strlen(base);
		} else 

		/*
		 * MODERATOR COMMANDS
		 */

		if (mod && strcasecmp(cmd, "getconfig") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoGetConfig(msg);
		    didSetConfig = 0;
		} else if (mod && strcasecmp(cmd, "setconfig") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    base = DoSetConfig(base+l, msg, arg, msg->ms_ExecScanOnly);
		    l = 0;
		    didSetConfig = 1;
		} else if (mod && strcasecmp(cmd, "getdisabled") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoGetList(msg, "GETDISABLED", '0');
		    didSetInfo = 0;
		} else if (mod && strcasecmp(cmd, "getinfo") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoGetInfo(msg);
		    didSetInfo = 0;
		} else if (mod && strcasecmp(cmd, "setinfo") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    base = DoSetInfo(base + l, msg, arg, msg->ms_ExecScanOnly);
		    l = 0;
		    didSetInfo = 1;
		} else if (mod && strcasecmp(cmd, "rundigest") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoRunDigest(msg->ms_List, msg);
		} else if (mod && strcasecmp(cmd, "setdigestvol") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoSetDigestVol(msg, strtol(arg, NULL, 0));
		} else if (mod && strcasecmp(cmd, "setdigestno") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoSetDigestNo(msg, strtol(arg, NULL, 0));
		} else if (mod && strcasecmp(cmd, "getlist") == 0) {
		    if (msg->ms_ExecScanOnly)
			msg->ms_ExecScanFlags |= EXSF_MODERATOR;
		    else
			DoGetList(msg, "GETLIST", 0);
		} else {
		    if (msg->ms_ExecScanOnly == 0)
			msgprintf(msg, "\tUnrecognized\n");
		}
	    }
	}
	base += l;
	if (*base == '\n')
	    ++base;
    }
    if (msg->ms_ExecScanOnly == 0) {
	if (didSetConfig)
	    DoGetConfig(msg);
	if (didSetInfo)
	    DoGetInfo(msg);
	if (didSubscribe > 0 && didSetInfo == 0 && !mod)
	    DoInfo(msg, mod);
    }
}

/*
 * subscribe [username]
 */

void
DoSubscribe(
    Msg *msg, 
    const char *arg, 
    int alen, 
    char digestFlag, 
    char aliasFlag, 
    char enableFlag,
    int mod
) {
    const char *email;
    int elen;
    int recipNotSender;
    char flags[32];

    if (alen == 0 || mod)
	recipNotSender = 0;
    else if (MsgFromCompare(msg, arg, alen) == 0)
	recipNotSender = 0;
    else
	recipNotSender = 1;

    {
	int n = 0;

	if (enableFlag)
	    flags[n++] = enableFlag;
	if (digestFlag)
	    flags[n++] = digestFlag;
	if (aliasFlag)
	    flags[n++] = aliasFlag;
	flags[n] = 0;
    }

    /*
     * extract full address / email
     */

    if (arg[0] == 0) {
        arg = findHeader(msg, "From:", 0, &alen, "");
        arg = parseAddress(arg, alen, &alen);
    }
    email = parseEMailFromAddress(arg, alen, &elen);
    if (alen > 128) {
	msgprintf(msg, "Unable to %s, your email address is too long,\n",
	    ((enableFlag == '+') ? "subscribe" : "unsubscribe")
	);
	msgprintf(msg, "128 chars max\n");
	return;
    }
    if (validEMailAddress(email, elen) < 0) {
	msgprintf(msg, "The email address '%*.*s' contains an illegal\n", elen, elen, email);
	msgprintf(msg, "character or is not in the user@host form\n");
	return;
    }
    if (arg && email && elen) {
	if (SetListFlags(msg->ms_List, email, elen, flags) == 0) {
	    msgprintf(msg, "\t%s <%*.*s> - successful%s\n",
		((enableFlag == '+') ? "subscribe" : "unsubscribe"),
		elen,
		elen,
		email,
		((enableFlag == '+' && digestFlag) ? " subscription to digest" : "")
	    );

	    /*
	     * Send the recipient an acknowledgement of the
	     * subscribe or unsubscribe.
	     */
	    if (recipNotSender) {
		Msg *msg2;
		
		msg2 = allocMsgFrom(
		    msg, 
		    arg, 
		    alen, 
		    ((enableFlag == '+') ? "Subscribe Acknowledgement" : "Unsubscribe Acknowledgement"),
		    "\n"
		);
		msg2->ms_BounceFlag |= BOUNCE_RETURN;
		msg2->ms_DeleteFlag = 1;
		msgprintf(msg2, "%*.*s has been %s %s\n",
		    elen,
		    elen,
		    email,
		    ((enableFlag == '+') ? "subscribed to" : "unsubscribed from"),
		    msg2->ms_ListName
		);
		disposeMsg(msg2);
	    }
	} else {
	    msgprintf(msg, "\t%s <%*.*s> - address not found or error\n",
		((enableFlag == '+') ? "subscribe" : "unsubscribe"),
		elen,
		elen,
		email
	    );
	}
    } else {
	if (arg)
	    msgprintf(msg, "Argument error in subscribe: %*.*s\n", alen, alen, arg);
    }
}

/*
 * AuthSubscribe() -
 *
 * Create a secondary message queue file who's state is set
 * to MSSTATE_AUTHSUBSCRIBE, then create a temporary message
 * to 'return' to the subscriber asking for verification.
 */

void
AuthSubscribe(
    Msg *msg, 
    const char *arg, 
    int alen, 
    char digestFlag, 
    char aliasFlag
) {
    Msg *msg2 = allocMsgFrom(msg, arg, alen, "Authorize Subscription", "%s %*.*s\nend\n", ((aliasFlag) ? "alias" : (digestFlag) ? "subscribe" : "subsingle"), alen, alen, arg);

    msg2->ms_BounceFlag |= BOUNCE_AUTHENTICATE;
    msg2->ms_IncludeCompleteMessage = 1;
    if (DumpLines("msgTPSub", msg2->ms_List, msg2, NULL, -1) == 0) {
	msgprintf(msg2, "Hello.  A subscription to %s has been requested for '%*.*s'\n", msg2->ms_ListName, alen, alen, arg);
	msgprintf(msg2, "The original command has been included below.  Your authorization will\n");
	msgprintf(msg2, "only effect this one subscribe.  If you do not wish to be subscribed,\n");
	msgprintf(msg2, "simply do not respond to this email\n");
    }
    disposeMsg(msg2);

    msgprintf(msg, "\tThird party subscribe verification sent to recipient\n");
}

/*
 * Retrieve configuration file
 */

void
DoGetConfig(Msg *msg)
{
    if (msg->ms_List) {
	FILE *fi;

	if ((fi = OpenListFile(msg->ms_ListName, "config", 0, -1)) != NULL) {
	    char buf[256];

	    msgprintf(msg, "\n");
	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		msgprintf(msg, "%s", buf);
	    }
	    msgprintf(msg, "\n");
	} else {
	    msgprintf(msg, "getconfig: unable to read config file\n");
	}
    } else {
	msgprintf(msg, "getconfig: command was not submitted to a list\n");
    }
}

/*
 * DoSetConfig() - store new configuration file
 */

const char *
DoSetConfig(const char *base, Msg *msg, char *arg, int scanOnly)
{
    if (msg->ms_List) {
	/*
	 * scan base for BEGIN CONFIG, END CONFIG
	 */
	int l;
	int state = 0;
	FILE *fo = NULL;

	if (scanOnly || (fo = OpenListFile(msg->ms_ListName, "config.tmp", O_CREAT, -1)) != NULL) {
	    while (state != 2 && *base) {
		for (l = 0; base[l] && base[l] != '\n'; ++l)
		    ;
		if (state == 0 && strncmp(base, "BEGIN CONFIG", 12) == 0) {
		    state = 1;
		} 
		if (state == 1) {
		    fwrite(base, 1, l, fo);
		    fputc('\n', fo);
		}
		if (state == 1 && strncmp(base, "END CONFIG", 10) == 0) {
		    state = 2;
		}
		if (base[l] == '\n')
		    ++l;
		base += l;
	    }
	    if (fo) {
		fflush(fo);
		if (ferror(fo)) {
		    msgprintf(msg, "Error writing new configuration file\n");
		    state = 0;
		}
		fclose(fo);
		if (state == 2) {
		    RenameListFile(msg->ms_ListName, "config", "config.old");
		    RenameListFile(msg->ms_ListName, "config.tmp", "config");
		    msgprintf(msg, "A new Configuration file has been installed\n\n");
		} else {
		    RemoveListFile(msg->ms_ListName, "config.tmp");
		    msgprintf(msg, "Format error trying to install configuration file\n");
		    msgprintf(msg, "A new configuration file has NOT been installed\n");
		}
	    }
	} else if (scanOnly == 0) {
	    msgprintf(msg, "Unable to create temporary file\n");
	}
    }
    return(base);
}

/*
 * Retrieve information file
 */

void
DoGetInfo(Msg *msg)
{
    if (msg->ms_List) {
	FILE *fi;

	if ((fi = OpenListFile(msg->ms_ListName, "info", 0, -1)) != NULL) {
	    char buf[256];

	    msgprintf(msg, "\n");
	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		msgprintf(msg, "%s", buf);
	    }
	    msgprintf(msg, "\n");
	} else {
	    msgprintf(msg, "getinfo: unable to read info file\n");
	    msgprintf(msg, "getinfo: info file may not exist\n\n");
	    msgprintf(msg, "BEGIN INFO %s\n", msg->ms_ListName);
	    msgprintf(msg, "END INFO %s\n", msg->ms_ListName);
	}
    } else {
	msgprintf(msg, "getinfo: command was not submitted to a list\n");
    }
}

/*
 * DoSetInfo() - store new configuration file
 */

const char *
DoSetInfo(const char *base, Msg *msg, char *arg, int scanOnly)
{
    if (msg->ms_List) {
	/*
	 * scan base for BEGIN INFO, END INFO
	 */
	int l;
	int state = 0;
	FILE *fo = NULL;

	if (scanOnly || (fo = OpenListFile(msg->ms_ListName, "info.tmp", O_CREAT, -1)) != NULL) {
	    while (state != 2 && *base) {
		for (l = 0; base[l] && base[l] != '\n'; ++l)
		    ;
		if (state == 0 && strncmp(base, "BEGIN INFO", 10) == 0) {
		    state = 1;
		} 
		if (state == 1) {
		    fwrite(base, 1, l, fo);
		    fputc('\n', fo);
		}
		if (state == 1 && strncmp(base, "END INFO", 8) == 0) {
		    state = 2;
		}
		if (base[l] == '\n')
		    ++l;
		base += l;
	    }
	    if (fo) {
		fflush(fo);
		if (ferror(fo)) {
		    msgprintf(msg, "Error writing new info file\n");
		    state = 0;
		}
		fclose(fo);
		if (state == 2) {
		    RenameListFile(msg->ms_ListName, "info", "info.old");
		    RenameListFile(msg->ms_ListName, "info.tmp", "info");
		    msgprintf(msg, "A new Info file has been installed\n\n");
		} else {
		    RemoveListFile(msg->ms_ListName, "info.tmp");
		    msgprintf(msg, "Format error trying to install info file\n");
		    msgprintf(msg, "A new info file has NOT been installed\n");
		}
	    }
	} else if (scanOnly == 0) {
	    msgprintf(msg, "Unable to create temporary file\n");
	}
    }
    return(base);
}


void
DoGetList(Msg *msg, const char *cmd, char c)
{
    FILE *fi;

    if (msg->ms_List) {
	msgprintf(msg, "BEGIN %s %s\n", cmd, msg->ms_ListName);
	if ((fi = OpenListFile(msg->ms_ListName, "list", 0, LOCK_SH)) != NULL) {
	    char buf[256];

	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		/*
		 * type flag restriction
		 */
		if (c) {
		    int i;

		    for (i = 0; buf[i] && buf[i]!=' ' && buf[i]!='\t'; ++i) {
			if (buf[i] == c)
			    break;
		    }
		    if (buf[i] != c)
			continue;
		}

		msgprintf(msg, "%s", buf);
	    }
	    fclose(fi);
	}
	msgprintf(msg, "END %s %s\n", cmd, msg->ms_ListName);
    } else {
	msgprintf(msg, "getlist: command was not submitted to a list\n");
    }
}

void
DoSetDigestVol(Msg *msg, int v)
{
    MailStats *stats;

    DoRunDigest(msg->ms_List, NULL);
    if ((stats = ReadMailStats(msg->ms_ListName)) != NULL) {
	stats->st_DigestVol = v;
	WriteMailStats(stats);
    }
    msgprintf(msg, "setdigestvol: Volume set to %d\n", v);
}

void
DoSetDigestNo(Msg *msg, int v)
{
    MailStats *stats;

    DoRunDigest(msg->ms_List, NULL);
    if ((stats = ReadMailStats(msg->ms_ListName)) != NULL) {
	stats->st_DigestNo = v;
	WriteMailStats(stats);
    }
    msgprintf(msg, "setdigestno: digest no. set to %d\n", v);
}

void
SaveDigest(Msg *msg, int msgFromMod)
{
    MailStats *stats;
    int maxDigestSize = GetListVal(msg->ms_List, "maxDigestSize", 0, 50*1024);
    int maxDigestMsgs = GetListVal(msg->ms_List, "maxDigestMsgs", 0, 10);
    int minDigestPoll = GetListVal(msg->ms_List,"minDigestPoll",0,2*60*60*24);
    int msgSize = msg->ms_Body.he_Bytes;
    int runIt = 0;
    FILE *fo;

    stats = ReadMailStats(msg->ms_ListName);

    /*
     * Try to be smart about longish postings
     */

    if (maxDigestSize && stats->st_DigestSize > 2 * maxDigestSize / 3 && 
	stats->st_DigestSize + msgSize >= maxDigestSize
    ) {
	DoRunDigest(msg->ms_List, NULL);
	/* stats will be adjusted */
    }

    /*
     * Re-set stat's file poll time if no messages in digest to
     * prevent an immediate run when no articles are in the digest.
     */
    if (stats->st_DigestMsgs == 0) {
	stats->st_DigestPoll = (long)time(NULL);
    }

    /*
     * Should we run digest after appending message ?
     */

    if (maxDigestSize && stats->st_DigestSize + msgSize >= maxDigestSize) {
	runIt = 1;
    }
    if (maxDigestMsgs && stats->st_DigestMsgs + 1 >= maxDigestMsgs) {
	runIt = 1;
    }
    if (minDigestPoll) {
	long dt = (long)time(NULL) - stats->st_DigestPoll;
	if (dt < -300 || dt > minDigestPoll)
	    runIt = 1;
    }

    if ((fo = OpenListFile(msg->ms_ListName, "digest.tmp", O_CREAT, -1)) == NULL) {
	/* LOG ERROR */
	return;
    }
    fseek(fo, 0L, 2);

    /*
     * From:, Subject:, message body
     */

    {
	const char *ptr;
	int len;

	fprintf(fo, 
	    "\n--------------- MESSAGE %s.v%03d.n%03d.%d ---------------\n\n",
	    msg->ms_ListName,
	    stats->st_DigestVol,
	    stats->st_DigestNo,
	    stats->st_DigestMsgs + 1
	);
	ptr = findHeader2(msg, msgFromMod, "From:", &len, "");
	fprintf(fo, "From: %*.*s\n", len, len, ptr);
	ptr = findHeader2(msg, msgFromMod, "Subject:", &len, "");
	ptr = stripPassword(ptr, len, &len);
	fprintf(fo, "Subject: %*.*s\n", len, len, ptr);
	ptr = findHeader2(msg, msgFromMod, "Date:", &len, "");
	fprintf(fo, "Date: %*.*s\n", len, len, ptr);
	CopyOptionalHeaders(msg, msgFromMod, -1, fo);
	fprintf(fo, "\n");
	ptr = stripPasswordFromBody(msg->ms_Body.he_Base, msg->ms_Body.he_Bytes, &len);

	fwrite(ptr, len, 1, fo);
	fprintf(fo, "\n");
    }

    fflush(fo);
    /* XXX check ferror */
    fclose(fo);

    /*
     * Index file
     */

    if ((fo = OpenListFile(msg->ms_ListName, "index.tmp", O_CREAT, -1)) != NULL) {
	const char *ptr;
	int len;
	int rhlen = 79 - 17;

	if (rhlen < 15)
	    rhlen = 15;

	fseek(fo, 0L, 2);
	fprintf(fo, "    %03d - ",
	    stats->st_DigestMsgs + 1
	);
	ptr = findHeader2(msg, msgFromMod, "From:", &len, "");
	if (len > 25)
	    len = 25;
	fprintf(fo, "%*.*s%*.*s - ", len, len, ptr, 25 - len, 25 - len, "");

	ptr = findHeader2(msg, msgFromMod, "Subject:", &len, "");
	ptr = stripPassword(ptr, len, &len);
	if (len > rhlen)
	    len = rhlen;
	fprintf(fo, "%*.*s\n", len, len, ptr);
	fclose(fo);
    }

    /*
     * Bump stats
     */

    ++stats->st_DigestMsgs;
    stats->st_DigestSize += msgSize; /* XXX include headers */
    stats->st_DigestNext = stats->st_DigestPoll + minDigestPoll;

    /*
     * run digest if we determined we should
     */

    if (runIt) {
	DoRunDigest(msg->ms_List, NULL);
    }
    WriteMailStats(stats);
}

/*
 * DoRunDigest() - do a digest run for this list.  The file
 * 		    digest is transmitted and renamed/deleted.
 */

void
DoRunDigest(MailList *list, Msg *msg)
{
    int pid;
    int fd;
    FILE *fi;
    FILE *fm;
    FILE *fs;
    const char *tp;
    MailStats *stats;
    int minDigestPoll = GetListVal(list, "minDigestPoll", 0, 2*60*60*24);
    int headerAfter = strcmp(GetListVar(list, "headerWhere", 0, "after"), "before") != 0;

    stats = ReadMailStats(list->ml_ListName);

    if ((fm = OpenListFile(list->ml_ListName, "digest.tmp", 0, -1)) == NULL) {
	WriteMailStats(stats);
	return;
    }
    fs = OpenListFile(list->ml_ListName, "index.tmp", 0, -1);

    fd = copenSendmail(&pid, list, "errors", 1);

    /*
     * User's From: and Subject:
     */

    cprintf(fd, "From: %s-errors@%s\n", list->ml_ListName, HostName);
    cprintf(fd, "Subject: Digest %s.v%03d.n%03d\n", 
	list->ml_ListName,
	stats->st_DigestVol,
	stats->st_DigestNo
    );

    /*
     * XXX what about looped digest postings ? (future)
     */

    cprintf(fd, "BestServHost: %s\n", HostName);

    /*
     * Our Sender, Errors-To, and To field (note: the To: line actually
     * causes the message to loop back to us, but the BestServHost:
     * header catches it and we throw it away).
     */

    cprintf(fd, "Sender: %s-errors@%s\n", list->ml_ListName, HostName);
    cprintf(fd, "Errors-To: %s-errors@%s\n", list->ml_ListName, HostName);
    if ((tp = GetListVar(list, "replyTo", 0, NULL)) != NULL && tp[0]) {
	cprintf(fd, "Reply-To: %s\n", tp);
    }
    cprintf(fd, "To: %s@%s\n", list->ml_ListName, HostName);

    /*
     * Generate Bcc: to the users on the list.  Is there a practical
     * limit or not ?
     */

    cprintf(fd, "Bcc:");

    if ((fi = OpenListFile(list->ml_ListName, "list", 0, LOCK_SH)) != NULL) {
	char buf[256];
	char *efull;
	char *flags;
	int n = 0;
	int digestOnly = (GetListVar(list, "digestOnly", 0, "N")[0] == 'Y');

	while (GetListFile(fi, buf, sizeof(buf), &flags, &efull) == 0) {
	    if ((digestOnly || strchr(flags, 'd') != NULL) && strchr(flags, 'a') == NULL) {
		int elen;
		const char *email;
		
		email = parseEMailFromAddress(efull, strlen(efull), &elen);
		if (email != NULL && elen > 0) {
		    if (n == 0) {
			cprintf(fd, " %*.*s", elen, elen, email);
		    } else {
			cprintf(fd, ",\n\t%*.*s", elen, elen, email);
		    }
		    ++n;
		}
	    }
	}
	fclose(fi);
    }
    cprintf(fd, "\n");	/* end of BCC */

    /*
     * blank line, then the message body
     */
    cprintf(fd, "\n");

    /*
     * Prepend headers
     */

    if (headerAfter == 0)
	DumpLines("header", list, NULL, NULL, fd);

    cprintf(fd, "\n-------------- BEGIN %s.v%03d.n%03d --------------\n\n",
	list->ml_ListName,
	stats->st_DigestVol,
	stats->st_DigestNo
    );

    /*
     * The index
     */

    if (fs) {
	char buf[1024];
	int n;

	while ((n = fread(buf, 1, sizeof(buf), fs)) > 0)
	    cwrite(fd, buf, n);
	fclose(fs);
    }

    /*
     * Postpend headers
     */

    if (headerAfter != 0)
	DumpLines("header", list, NULL, NULL, fd);

    /*
     * The digest body
     */

    {
	char buf[1024];
	int n;

	while ((n = fread(buf, 1, sizeof(buf), fm)) > 0)
	    cwrite(fd, buf, n);
    }

    fclose(fm);

    cprintf(fd, "\n--------------- END %s.v%03d.n%03d ---------------\n\n",
	list->ml_ListName,
	stats->st_DigestVol,
	stats->st_DigestNo
    );

    /*
     * footer's
     */
    DumpLines("footer", list, NULL, NULL, fd);

    cclose(fd, pid);

    if (msg) {
	if (stats->st_DigestMsgs) {
	    msgprintf(msg, "rundigest - digest run %s.v%03d.n%03d (%d msgs, %d bytes)\n",
		list->ml_ListName,
		stats->st_DigestVol,
		stats->st_DigestNo,
		stats->st_DigestMsgs,
		stats->st_DigestSize
	    );
	} else {
	    msgprintf(msg, "rundigest - no messages were pending\n");
	}
    }

    ++stats->st_DigestNo;
    stats->st_DigestSize = 0;
    stats->st_DigestMsgs = 0;
    stats->st_DigestPoll = (long)time(NULL);
    stats->st_DigestNext = stats->st_DigestPoll + minDigestPoll;
    RemoveListFile(list->ml_ListName, "index");
    RemoveListFile(list->ml_ListName, "digest");
    RenameListFile(list->ml_ListName, "digest.tmp", "digest");
    RenameListFile(list->ml_ListName, "index.tmp", "index");
    WriteMailStats(stats);
}

void 
DoHelp(Msg *msg, int mod)
{
    FILE *fi;
    char path[256];

    msg->ms_ErrorMsg = "HELP RESPONSE";

    msgprintf(msg, "HELP for BESTSERV list server\n\n");

    snprintf(path, sizeof(path), "%s/msgs/help.gen", BaseDir);

    if ((fi = fopen(path, "r")) != NULL) {
	char buf[256];

	msgprintf(msg, "BEGIN HELP GENERAL\n\n");

	while (fgets(buf, sizeof(buf), fi) != NULL) {
	    msgprintf(msg, "%s", buf);
	}
	fclose(fi);
	msgprintf(msg, "END HELP GENERAL\n\n");
    }

    if (mod) {
	snprintf(path, sizeof(path), "%s/msgs/help.mod", BaseDir);

	if ((fi = fopen(path, "r")) != NULL) {
	    char buf[256];

	    msgprintf(msg, "BEGIN HELP MODERATOR\n\n");

	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		msgprintf(msg, "%s", buf);
	    }
	    fclose(fi);

	    msgprintf(msg, "END HELP MODERATOR\n\n");
	}
    }
}

void
DoInfo(Msg *msg, int mod)
{
    FILE *fi;

    msg->ms_ErrorMsg = "INFO RESPONSE";

    msgprintf(msg, "\n----- INFORMATION FOR LIST %s -----\n\n", msg->ms_ListName);

    /*
     * info lines in config file
     */

    DumpLines("info", msg->ms_List, msg, NULL, -1);

    /*
     * info file
     */

    if ((fi = OpenListFile(msg->ms_ListName, "info", 0, -1)) != NULL) {
	char buf[256];

	msgprintf(msg, "\n");
	while (fgets(buf, sizeof(buf), fi) != NULL) {
	    msgprintf(msg, "%s", buf);
	}
	fclose(fi);
    }
}

void
CopyOptionalHeaders(Msg *msg, int msgFromMod, int fd, FILE *fo)
{
    static char *OptHead[] = {
	"MIME-Version:",
	"Content-Type:",
	"Content-Length:",
	"Content-Transfer-Encoding:",
	"Content-Disposition:",
	"References:",
	"In-Reply-To:",
	NULL
    };
    int i;

    for (i = 0; OptHead[i]; ++i) {
	const char *p;
	int l;

	p = findHeader2(msg, msgFromMod, OptHead[i], &l, NULL);
	if (p && l < 512) {
	    if (fd >= 0)
		cprintf(fd, "%s %*.*s\n", OptHead[i], l, l, p);
	    else if (fo)
		fprintf(fo, "%s %*.*s\n", OptHead[i], l, l, p);
	}
    }
}

int
CheckFromNNTP(Msg *msg)
{
    int len;

    if (findHeader(msg, "NNTP-Posting-Host:", 0, &len, NULL))
	return(1);
    return(0);
}

