
/*
 * MSG.C
 */

#include "defs.h"

Prototype void RunMsg(Msg *msg);
Prototype Msg *LoadMessageById(Msg *omsg, const char *id);
Prototype void ChangeMsgId(Msg *msg, char newState);
Prototype Msg *collectMsg(int, const char *, const char *, const char *, int qfo);
Prototype Msg * allocMsgFrom(Msg *omsg, const char *from, int fromLen, const char *sub, const char *ctl, ...);
Prototype void disposeMsg(Msg *msg);
Prototype const char *findHeader(Msg *, const char *, int, int *, const char *);
Prototype int checkHeader(Msg *msg, const char *hdr, const char *substr);
Prototype const char *findHeader2(Msg *, int msgFromMod, const char *, int *, const char *);
Prototype const char *findHeaderV(Msg *msg, const char *hdr, int idx, int *plen, const char *def);
Prototype const char *parseAddress(const char *field, int len, int *plen);
Prototype const char *parseEMailFromAddress(const char *field, int len, int *plen);
Prototype const char *stripPassword(const char *field, int len, int *plen);
Prototype const char *stripPasswordFromBody(const char *field, int len, int *plen);
Prototype void msgprintf(Msg *msg, const char *ctl, ...);
Prototype const char *ModeratorEMail(Msg *msg, int *plen);
Prototype int MsgFromModerator(Msg *msg, int *pnum);
Prototype int MsgFromCompare(Msg *msg, const char *from, int fromLen);

int createQueueFile(char **pqf, int *poff);
void dumpMsg(Msg *msg, int fd, int incl);

/*
 * RunMsg() - run message through subsystem
 */

void 
RunMsg(Msg *msg)
{
    if (msg->ms_SubSystem == NULL) {
	msg->ms_DeleteFlag = 1;
	msg->ms_ErrorFlag = 1;
	msg->ms_ErrorMsg = "NULL Subsystem";
	return;
    }
    if (msg->ms_State == 'P') {
	msg->ms_DeleteFlag = 1;
	msg->ms_ErrorFlag = 1;
	msg->ms_ErrorMsg = "Partially Instantiated Queue File";
	return;
    }

    /*
     * Look through BestServHost headers, ignore message if we've already
     * seen it.  This allows multiple BestServ's to be bi-directionally
     * networked without creating loops.
     */
    
    {
	const char *host;
	int len;
	int n = 0;

	while ((host=findHeader(msg, "BestServHost:", n, &len, NULL)) != NULL ){
	    if (strlen(HostName)==len && strncasecmp(host,HostName,len) == 0) {
		msg->ms_DeleteFlag = 1;
		msg->ms_ErrorMsg = "Message Loop, message thrown away";
		return;
	    }
	    ++n;
	}
    }

    if (strcmp(msg->ms_SubSystem, "post") == 0) {
	PostProc(msg);
    } else if (strcmp(msg->ms_SubSystem, "digest") == 0) {
	PostProc(msg);
    } else if (strcmp(msg->ms_SubSystem, "request") == 0) {
	RequestProc(msg);
    } else if (strcmp(msg->ms_SubSystem, "command") == 0) {
	CommandProc(msg);
    } else if (strcmp(msg->ms_SubSystem, "response") == 0) {
	ResponseProc(msg);
    } else if (strcmp(msg->ms_SubSystem, "errors") == 0) {
	ErrorsProc(msg);
    } else if (strcmp(msg->ms_SubSystem, "owner") == 0) {
	msg->ms_BounceFlag |= BOUNCE_FWD_MODERATOR;
	msg->ms_DeleteFlag = 1;
    } else {
	/* do not delete */
	msg->ms_ErrorFlag = 1;
	msg->ms_ErrorMsg = "Unknown Subsystem";
	msg->ms_DeleteFlag = 1;
    }
}

/*
 * LOADMESSAGEBYID()
 *
 * ID format: <x%08lx.%ld.randomid>
 *	       \________/
 *		queue file
 *		name
 */

Msg *
LoadMessageById(Msg *omsg, const char *id)
{
    Msg *msg = NULL;

    if (id && *id == '<') {
	int j;
	int k = 0;
	int qfo = 0;
	char buf[256];

	for (++id, j = 0; id[j] && id[j] != '>'; ++j) {
	    if (id[j] == '.') {
		if (++k == 2)
		    break;
	    }
	}
	if (k == 2 && j < 64) {	/* k positioned at second '.' */
	    snprintf(buf, sizeof(buf), "%s/spool/%*.*s",
		BaseDir,
		j, j, id
	    );
	    qfo = strlen(BaseDir) + 7;
	    if ((msg = collectMsg(-1, buf, NULL, NULL, qfo)) != NULL) {
		if (strncmp(msg->ms_Id, id, strlen(msg->ms_Id)) != 0) {
		    msg->ms_ErrorMsg = "ID mismatch on verify response";
		    disposeMsg(msg);
		    msg = NULL;
		}
	    }
	}
    }
    return(msg);
}

void 
ChangeMsgId(Msg *msg, char newState)
{
    if (msg->ms_IdOff) {
	char buf[256];

	snprintf(buf, sizeof(buf), "I%s.%s\n",
	    msg->ms_QueueFile + msg->ms_QFO,
	    MakeId(msg)
	);
	msg->ms_State = newState;
	lseek(msg->ms_Fd, 1, 0);
	write(msg->ms_Fd, &msg->ms_State, 1);
	lseek(msg->ms_Fd, msg->ms_IdOff, 0);
	write(msg->ms_Fd, buf, strlen(buf));
	freestr(&msg->ms_Id);
	buf[strlen(buf) - 1] = 0;
	msg->ms_Id = dupstr(buf + 1);
    }
}


Msg *
collectMsg(
    int fd, 
    const char *queueFile, 
    const char *listName, 
    const char *subSystem,
    int qfo
) {
    Msg *msg = zalloc(sizeof(Msg));

    msg->ms_Fd = -1;

    if (queueFile == NULL) {
	if ((msg->ms_Fd = createQueueFile(&msg->ms_QueueFile, &qfo)) >= 0) {
	    int n;
	    char buf[4096];

	    msg->ms_DeleteFlag = 1;
	    msg->ms_QFO = qfo;

	    snprintf(buf, sizeof(buf), "CP\nL%s\nS%s\nI%s.%s\n\n",
		((listName) ? listName : "-"),
		subSystem,
		msg->ms_QueueFile + qfo,
		MakeId(msg)
	    );

	    if (write(msg->ms_Fd, buf, strlen(buf)) != strlen(buf))
		msg->ms_ErrorFlag = 1;
	    while ((n = read(fd, buf, sizeof(buf))) > 0) {
		if (write(msg->ms_Fd, buf, n) != n) {
		    msg->ms_ErrorFlag = 1;
		    break;
		}
	    }
	    if (fsync(msg->ms_Fd) < 0) {
		msg->ms_ErrorFlag = 1;
	    } else {
		lseek(msg->ms_Fd, 1, 0);
		write(msg->ms_Fd, "A", 1);
	    }
	} else {
	    msg->ms_ErrorFlag = 1;
	    msg->ms_ErrorMsg = "Unable to create queue file";
	}
    } else {
	msg->ms_QueueFile = dupstr(queueFile);
	msg->ms_QFO = qfo;
	if ((msg->ms_Fd = open(queueFile, O_RDWR)) >= 0) {
	    if (flock(msg->ms_Fd, LOCK_EX | LOCK_NB) < 0) {
		msg->ms_ErrorFlag = 1;
	    }
	} else {
	    msg->ms_ErrorFlag = 1;
	}
    }

    /*
     * If no problems, then msg->ms_QueueFile is valid and
     * msg->ms_Fd is valid and locked.  Map the message into
     * memory.
     */

    if (msg && msg->ms_ErrorFlag == 0) {
	msg->ms_Size = lseek(msg->ms_Fd, 0L, 2);
	msg->ms_Map = mmap(
	    (caddr_t)0,
	    msg->ms_Size + 1,
	    PROT_READ,
	    MAP_SHARED,
	    msg->ms_Fd,
	    0
	);
	if (msg->ms_Map == NULL || msg->ms_Map == (char *)-1)
	    fatal(errno, "mmap(%s,%d)", msg->ms_QueueFile, msg->ms_Size + 1);
    }

    /*
     * Still ok, scan message headers in two passes
     */

    if (msg && msg->ms_ErrorFlag == 0) {
	int pass;
	const char *base = NULL;

	for (pass = 1; pass <= 2; ++pass) {
	    int n = 0;

	    base = msg->ms_Map;

	    /*
	     * Scan control commands until we see a blank line
	     */

	    while (*base && *base != '\n') {
		const char *ptr;

		if ((ptr = strchr(base, '\n')) != NULL) {
		    if (pass == 2) {
			int l = ptr - base - 1;

			switch(*base) {
			case 'C':
			    if (base[1] == '\n' || base[1] == 0)
				msg->ms_State = 'P';
			    else
				msg->ms_State = base[1];
			    break;
			case 'L':
			    msg->ms_ListName = dupnstr(base + 1, l);
			    break;
			case 'S':
			    msg->ms_SubSystem = dupnstr(base + 1, l);
			    break;
			case 'I':
			    msg->ms_Id = dupnstr(base + 1, l);
			    msg->ms_IdOff = base - msg->ms_Map;
			    break;
			}
		    }
		    ++ptr;
		} else {
		    ptr = base + strlen(base);
		}
		base = ptr;
	    }

	    if (*base == '\n')
		++base;
	    else
		msg->ms_State = 'P';

	    /*
	     * Scan headers until we see a blank line
	     */

	    while (*base && *base != '\n') {
		const char *ptr;

		/*
		 * Locate end of header.  A header may contain additional
		 * lines if it begins with a space or tab.
		 */

		if ((ptr = strchr(base, '\n')) != NULL) {
		    for (++ptr; *ptr == ' ' || *ptr == '\t'; ++ptr) {
			if ((ptr = strchr(ptr, '\n')) == NULL) 
			    break;
		    }
		}
		if (ptr == NULL) {
		    ptr = base + strlen(base);
		}

		if (pass == 2) {
		    msg->ms_Headers[n].he_Base = base;
		    msg->ms_Headers[n].he_Bytes = ptr - base - 1;
		}
		++n;
		base = ptr;
	    }
	    if (pass == 1) {
		msg->ms_NumHead = n;
		msg->ms_Headers = zalloc(sizeof(Header) * n);
	    }

	    /*
	     * Skip past header/body demark
	     */

	    if (*base == '\n')
		++base;
	}

	/*
	 * Store location of message body
	 */
	msg->ms_Body.he_Base  = base;
	msg->ms_Body.he_Bytes = msg->ms_Size - (base - msg->ms_Map);
    }

    if (msg->ms_Id == NULL) {
	msg->ms_ErrorFlag = 1;
	msg->ms_ErrorMsg = "Unable to locate ID";
    }

    /*
     * Load list information
     */

    if (msg && msg->ms_ListName && msg->ms_ErrorFlag == 0) {
	msg->ms_List = GetMailList(msg->ms_ListName, 0);
    }

    /*
     * return message or dispose and return NULL
     */

    if (msg->ms_ErrorFlag) {
	disposeMsg(msg);
	msg = NULL;
    }
    return(msg);
}

/*
 * allocMsgFrom() - allocate a message based on an existing message.
 *
 */

Msg *
allocMsgFrom(Msg *omsg, const char *from, int fromLen, const char *sub, const char *ctl, ...)
{
    int fd;
    int qfo;
    char *qfile;
    Msg *msg = NULL;

    if ((fd = createQueueFile(&qfile, &qfo)) >= 0) {
	FILE *fo = fdopen(fd, "r+");

	fprintf(fo, "CG\nL%s\nS%s\nI%s.%s\n\n",
	    omsg->ms_ListName,
	    "request",
	    qfile + qfo,
	    MakeId(omsg)
	);
	fprintf(fo, "From: %*.*s\n", fromLen, fromLen, from);
	fprintf(fo, "Subject: %s\n", sub);
	fprintf(fo, "\n");
	{
	    va_list va;
	    va_start(va, ctl);
	    vfprintf(fo, ctl, va);
	    va_end(va);
	}
	fclose(fo);

	msg = collectMsg(-1, qfile, NULL, NULL, qfo);
	msg->ms_ErrorMsg = sub;
	msg->ms_OriginalMsg = omsg;
	freestr(&qfile);
    }
    return(msg);
}

void
disposeMsg(Msg *msg)
{
    if (msg) {
	if (msg->ms_BounceFlag & BOUNCE_RETURN) {
	    const char *err = (msg->ms_ErrorMsg) ? msg->ms_ErrorMsg : "List Server Error";
	    const char *ptr;
	    int len;
	    int fd = -1;
	    int pid = -1;

	    if ((ptr = findHeaderV(msg, "From:", 0, &len, NULL)) != NULL) {
		fd = copenSendmail(&pid, msg->ms_List, "errors", 0);
		cprintf(fd, "From: %s-errors@%s\n", msg->ms_ListName, HostName);
		cprintf(fd,"Sender: %s-errors@%s\n",msg->ms_ListName, HostName);
		cprintf(fd, "To: %*.*s\n", len, len, ptr);
		cprintf(fd, "Subject: %s\n", err);
		cprintf(fd, "\n");
		cwrite(fd, msg->ms_ReturnBuf, msg->ms_RBLen);
		dumpMsg(msg, fd, 1);
		cclose(fd, pid);
	    }
	}

	if (msg->ms_BounceFlag & BOUNCE_AUTHENTICATE) {
	    const char *ptr;
	    int len;
	    int fd = -1;
	    int pid = -1;

	    if ((ptr = findHeaderV(msg, "From:", 0, &len, NULL)) != NULL) {
		fd = copenSendmail(&pid, msg->ms_List, "errors", 0);
		cprintf(fd, "From: %s-response@%s\n", msg->ms_ListName, HostName);
		cprintf(fd, "Sender: %s-response@%s\n", msg->ms_ListName, HostName);
		cprintf(fd, "Reply-To: %s-response@%s\n", msg->ms_ListName, HostName);
		cprintf(fd, "To: %*.*s\n", len, len, ptr);
		cprintf(fd, "Subject: List Auth Request ID=<%s> REJECT\n", msg->ms_Id);
		cprintf(fd, "\n");
		cwrite(fd, msg->ms_ReturnBuf, msg->ms_RBLen);
		dumpMsg(msg, fd, 1);
		cclose(fd, pid);
	    }
	}

	if (msg->ms_BounceFlag & BOUNCE_AUTHMODERATOR) {
	    const char *ptr;
	    int len;
	    int fd = -1;
	    int pid = -1;

	    fd = copenSendmail(&pid, msg->ms_List, "errors", 0);
	    cprintf(fd, "From: %s-response@%s\n", msg->ms_ListName, HostName);
	    cprintf(fd, "Sender: %s-response@%s\n", msg->ms_ListName, HostName);
	    cprintf(fd, "Reply-To: %s-response@%s\n", msg->ms_ListName, HostName);
	    ptr = ModeratorEMail(msg, &len);
	    cprintf(fd, "To: %*.*s\n", len, len, ptr);
	    cprintf(fd, "Subject: List Auth Request ID=<%s> REJECT\n", msg->ms_Id);
	    cprintf(fd, "\n");
	    cwrite(fd, msg->ms_ReturnBuf, msg->ms_RBLen);
	    dumpMsg(msg, fd, 1);
	    cclose(fd, pid);
	}

	if (msg->ms_BounceFlag & (BOUNCE_CCMODERATOR | BOUNCE_FWD_MODERATOR)) {
	    const char *ptr;
	    int len;
	    int fd = -1;
	    int pid = -1;
	    int incl = (msg->ms_BounceFlag & BOUNCE_CCMODERATOR) ? 1 : 0;

	    fd = copenSendmail(&pid, msg->ms_List, "errors", 0);
	    cprintf(fd, "From: %s-command@%s\n", msg->ms_ListName, HostName);
	    cprintf(fd, "Sender: %s-command@%s\n", msg->ms_ListName, HostName);
	    cprintf(fd, "Reply-To: %s-command@%s\n", msg->ms_ListName, HostName);

	    ptr = ModeratorEMail(msg, &len);
	    cprintf(fd, "To: %*.*s\n", len, len, ptr);
	    if (msg->ms_BounceFlag & BOUNCE_CCMODERATOR)
		cprintf(fd, "Subject: Moderator CC - %s\n", msg->ms_ListName);
	    else
		cprintf(fd, "Subject: EMail from list user - %s\n", msg->ms_ListName);
	    cprintf(fd, "\n");
	    if ((msg->ms_BounceFlag & BOUNCE_CCMODERATOR) != 0) {
		cprintf(fd, "(This is an FYI, No further action is required by the moderator)\n");
		cprintf(fd, "\n");
	    }
	    cwrite(fd, msg->ms_ReturnBuf, msg->ms_RBLen);
	    dumpMsg(msg, fd, incl);
	    cclose(fd, pid);
	}

	if (msg->ms_BounceFlag & BOUNCE_CCERRORS) {
	    const char *ptr;

	    ptr = GetListVar(msg->ms_List, "errorsCC", 0, NULL);

	    if (ptr != NULL && strcasecmp(ptr, "moderator") == 0)
		ptr = GetListVar(msg->ms_List, "moderator", 1, NULL);

	    if (ptr != NULL && *ptr != 0) {
		int fd = -1;
		int pid = -1;

		fd = copenSendmail(&pid, msg->ms_List, "owner", 0);
		cprintf(fd, "From: %s-owner@%s\n", msg->ms_ListName, HostName);
		cprintf(fd, "Sender: %s-owner@%s\n", msg->ms_ListName, HostName);
		cprintf(fd, "To: %s\n", ptr);
		cprintf(fd, "Subject: Bounced Message\n");
		cprintf(fd, "\n");
		cwrite(fd, msg->ms_ReturnBuf, msg->ms_RBLen);
		dumpMsg(msg, fd, 1);
		cclose(fd, pid);
	    }
	}

	if (msg->ms_ErrorFlag) {
	    printf("ERROR!\n");
	}
	if (msg->ms_ErrorMsg) {
	    printf("MSG: %s\n", msg->ms_ErrorMsg);
	}
	if (msg->ms_List) {
	    CloseMailList(msg->ms_List);
	    msg->ms_List = NULL;
	}

	if (msg->ms_DeleteFlag && msg->ms_QueueFile != NULL)
	    remove(msg->ms_QueueFile);

	freestr(&msg->ms_ListName);
	freestr(&msg->ms_SubSystem);
	freestr(&msg->ms_QueueFile);
	freestr(&msg->ms_Id);
	if (msg->ms_Map != NULL && msg->ms_Map != (char *)-1) {
	    munmap((void *)msg->ms_Map, msg->ms_Size + 1);
	    msg->ms_Size = 0;
	}
	msg->ms_Map = NULL;
	if (msg->ms_Fd >= 0) {
	    close(msg->ms_Fd);
	    msg->ms_Fd = -1;
	}
	if (msg->ms_ReturnBuf) {
	    free(msg->ms_ReturnBuf);
	    msg->ms_ReturnBuf = 0;
	    msg->ms_RBLen = 0;
	}
	zfree(msg->ms_Headers, sizeof(Header) * msg->ms_NumHead);
	zfree(msg, sizeof(Msg));
    }
}

void
dumpMsg(Msg *msg, int fd, int incl)
{
    int maxHead = 4096;
    int maxBody = 1024;

    if (msg->ms_IncludeCompleteMessage || incl == 0) {
	maxHead = 0x7FFFFFFF;
	maxBody = 0x7FFFFFFF;
    }

    /*
     * We dump the ORIGINAL message, not a derived message.
     */

    if (msg->ms_OriginalMsg)
	msg = msg->ms_OriginalMsg;

    /*
     * include (part of) the original message
     */

    if (fd >= 0) {
	int i;
	int bytes = 0;
	int dummyLen;

	/*
	 * up to 4K worth of headers
	 */
	cprintf(fd, "\n");
	for (i = 0; i < msg->ms_NumHead && bytes < maxHead; ++i) {
	    Header *head = &msg->ms_Headers[i];
	    int n = head->he_Bytes;

	    /*
	     * never include an approved: header in any response, even
	     * to a moderator.
	     */

	    if (strncasecmp(head->he_Base, "approved:", 9) == 0)
		continue;

	    if (bytes + n > maxHead)
		n = maxHead - bytes;
	    cwrite(fd, ":", 1);
	    cwrite(fd, head->he_Base, n);
	    cwrite(fd, "\n", 1);
	    bytes += n;
	}
	if (bytes == maxHead)
	    cprintf(fd, ":...\n");
	cprintf(fd, "%s\n", ((incl) ? ":" : ""));
	/*
	 * up to 1K worth of body.  Never include an embedded password.
	 */
	bytes = stripPasswordFromBody(msg->ms_Body.he_Base, msg->ms_Body.he_Bytes, &dummyLen) - msg->ms_Body.he_Base;

	for (; bytes < msg->ms_Body.he_Bytes && bytes < maxBody; ) {
	    int n = bytes;
	    while (n < maxBody && n < msg->ms_Body.he_Bytes && msg->ms_Body.he_Base[n] != '\n')
		++n;
	    if (msg->ms_Body.he_Base[n] == '\n')
		++n;
	    n -= bytes;
	    if (bytes + n > maxBody)
		n = maxBody - bytes;
	    if (incl)
		cwrite(fd, ":", 1);
	    cwrite(fd, msg->ms_Body.he_Base + bytes, n);
	    bytes += n;
	}
	if (bytes >= maxBody)
	    cprintf(fd, "\n:...\n");
    }
}

/*
 * FINDHEADER() - find a mail header
 *
 * Locate the specified mail header, such as "From:", and return the
 * contents of the field with pre and post whitespace removed.  The
 * result is NOT zero terminated.
 */

const char *
findHeader(Msg *msg, const char *hdr, int idx, int *plen, const char *def)
{
    int l = strlen(hdr);
    int i;

    if (def)
	*plen = strlen(def);
    else
	*plen = 0;

    for (i = 0; i < msg->ms_NumHead; ++i) {
	Header *h = &msg->ms_Headers[i];

	if (h->he_Bytes >= l && strncasecmp(hdr, h->he_Base, l) == 0) {
	    int ilen;
	    int jlen;

	    /*
	     * skip leading spaces/tabs
	     */
	    for (ilen = l; ilen < h->he_Bytes; ++ilen) {
		if (h->he_Base[ilen] != ' ' && h->he_Base[ilen] != '\t')
		    break;
	    }

	    /*
	     * skip trailing spaces/tabs
	     */
	    for (jlen = h->he_Bytes - 1; jlen >= ilen; --jlen) {
		if (h->he_Base[jlen] != ' ' && h->he_Base[jlen] != '\t')
		    break;
	    }
	    ++jlen;

	    if (--idx < 0) {
		def = h->he_Base + ilen;
		*plen = jlen - ilen;
		break;
	    }
	}
    }
    return(def);
}

/*
 * Same as findHeader(), but if the moderator is authenticated allows
 * body header overrides.
 */

const char *
findHeader2(Msg *msg, int msgFromMod, const char *hdr, int *plen, const char *def)
{
    /*
     * normal findHeader()
     */
    def = findHeader(msg, hdr, 0, plen, def);

    /*
     * locate header in body
     */

    if (msgFromMod == FROM_MODERATOR_AUTH) {
	int l = strlen(hdr);
	int initial = 1;
	const char *line;
	const char *next = NULL;

	for (line = msg->ms_Body.he_Base; line && *line; line = next) {
	    if ((next = strchr(line, '\n')) == NULL)
		next = line + strlen(line);
	    else
		++next;
	    if (*line == '\n') {
		if (initial)
		    continue;
		break;
	    }
	    initial = 0;
	    if (next - line >= l && strncasecmp(hdr, line, l) == 0) {
		int hlen = next - line - l;

		def = line + l;
		if (hlen > 0 && def[hlen-1] == '\n')
		    --hlen;
		while (hlen>0 && (def[hlen-1] == ' ' || def[hlen-1] == '\t'))
		    --hlen;
		while (hlen > 0 && (*def == ' ' || *def == '\t')) {
		    --hlen;
		    ++def;
		}
		*plen = hlen;
		break;
	    }
	}
    }
    return(def);
}

/*
 * findHeaderV() - locate header, but return default if the header
 * is too long or if it contains a semicolon or pipe symbol.
 */

const char *
findHeaderV(Msg *msg, const char *hdr, int idx, int *plen, const char *def)
{
    const char *ptr = findHeader(msg, hdr, idx, plen, def);
    int useDef = 0;

    if (*plen > 128) {
	useDef = 1;
    } else {
	int i;

	for (i = 0; i < *plen; ++i) {
	    switch(ptr[i]) {
	    case '|':
	    case ';':
		useDef = 1;
		break;
	    default:
		break;
	    }
	}
    }
    if (useDef) {
	ptr = def;
	*plen = (ptr) ? strlen(ptr) : 0;
    }
    return(ptr);
}

/*
 * Determine if the substring appears anywhere in the header.  Return the
 * header index or -1 if the string could not be found.
 */
int
checkHeader(Msg *msg, const char *hdr, const char *substr)
{
    int n;
    int len;
    const char *str;
    char *copy;

    if (len > 16384)
	return(-1);

    for (n = 0; (str = findHeader(msg, hdr, n, &len, NULL)) != NULL; ++n) {
	copy = malloc(len + 1);
	bcopy(str, copy, len);
	copy[len] = 0;
	if (strcasestr(copy, substr)) {
	    free(copy);
	    return(n);
	}
	free(copy);
    }
    return(-1);
}

/*
 * MODERATOREMAIL() - return email address of moderator
 *
 * Returns the email address of the moderator.  If the message
 * came FROM a moderator, and the From: line matches one of the
 * moderators, then the email address of that moderator is
 * returned.  Otherwise the email address of the principle
 * moderator is returned.
 */

const char *
ModeratorEMail(Msg *msg, int *plen)
{
    int n;
    int r = MsgFromModerator(msg, &n);
    const char *str = NULL;

    switch (r) {
    case FROM_MODERATOR_AUTH:
	/*
	 * If the person has a valid password, the moderator
	 * address is wherever he is coming from.
	 */
	str = findHeader(msg, "From:", 0, plen, NULL);
	if (str != NULL && *plen)
	    break;
	/* fall through */
    case FROM_MODERATOR:
	/*
	 * If the person is coming from a registered moderator
	 * address we send the mail to moderatorN
	 */
	str = GetListVar(msg->ms_List, "moderator", n, NULL);
	*plen = strlen(str);
	break;
    default:
	/*
	 * If the person is coming from an unknown address, we
	 * send the mail to moderator1.
	 */
	str = GetListVar(msg->ms_List, "moderator", 1, NULL);
	*plen = strlen(str);
	break;
    }
    if (str == NULL) {
	str = "bestserv-owner";
	*plen = strlen(str);
    }
    return(str);
}

/*
 * MSGFROMMODERATOR() - returns a status value indicating
 *			whether the sender was a moderator
 *			or not.
 *
 *			A secondary return value is the
 *			moderatorXXX field number.
 *
 *	FROM_UNKNOWN		- message not from moderator
 *	FROM_MODERATOR		- message apparently from a moderator
 *	FROM_MODERATOR_AUTH	- message included valid password
 */

int
MsgFromModerator(Msg *msg, int *pnum)
{
    const char *email;
    int elen;

    if (pnum)
	*pnum = 1;

    email = findHeader(msg, "From:", 0, &elen, "");
    email = parseEMailFromAddress(email, elen, &elen);

    if (elen != 0) {
	const char *modEmail;
	int n;

	for (n = 1; (modEmail = GetListVar(msg->ms_List, "moderator", n, NULL)) != NULL; ++n) {
	    if (modEmail[0]) {
		if (strlen(modEmail) == elen &&
		    strncasecmp(modEmail, email, elen) == 0
		) {
		    break;
		}
	    }
	}
	{
	    const char *pas;
	    const char *p2;
	    int len;

	    if (pnum)
		*pnum = n;

	    /*
	     * check for PASSWORD=<pw> in subject
	     */

	    pas = findHeader(msg, "Subject:", 0, &len, "");
	    if ((p2 = strstr(pas, "PASSWORD=<")) != NULL && p2 - pas < len) {
		const char *p3;

		p2 += 10;
		if ((p3 = strchr(p2, '>')) != NULL && p3 - pas < len) {
		    const char *lpas;
		    if ((lpas = GetListVar(msg->ms_List, "password", 0, ""))[0]) {
			if (strlen(lpas) == p3 - p2 &&
			    strncmp(lpas, p2, p3 - p2) == 0
			) {
			    return(FROM_MODERATOR_AUTH);
			}
		    }
		}
	    }

	    /*
	     * check for password in Approved: header
	     */

	    if ((pas = findHeader(msg, "Approved:", 0, &len, NULL)) != NULL) {
		const char *lpas;

		if ((lpas = GetListVar(msg->ms_List, "password", 0, ""))[0]) {
		    if (strlen(lpas) == len && strncmp(lpas, pas, len) == 0) {
			return(FROM_MODERATOR_AUTH);
		    }
		}
	    }

	    /*
	     * check for password in Body of message
	     */

	    pas = msg->ms_Body.he_Base;

	    /*
	     * Skip initial white space
	     */
	    while (*pas == ' ' || *pas == '\t' || *pas == '\n' || *pas == '\r')
		++pas;

	    /*
	     * Does body begin with 'approved:' ?
	     */
	    if (strncasecmp(pas, "approved:", 9) == 0) {
		/*
		 * skip approved keyword and possible colon (be generous)
		 */
		while (*pas && *pas != ' ' && *pas != '\n' && *pas != '\t')
		    ++pas;

		/*
		 * skip white space
		 */
		while (*pas == ' ' || *pas == '\t')
		    ++pas;

		/*
		 * Locate end of approved line, skip whitspace going backwards
		 */
		if ((p2 = strchr(pas, '\n')) != NULL) {
		    const char *lpas;

		    while (p2 >= pas && (*p2 == '\r' || *p2 == '\n' || *p2 == ' ' || *p2 == '\t'))
			--p2;
		    ++p2;
		    if ((lpas = GetListVar(msg->ms_List, "password", 0, ""))[0]) {
			if (strlen(lpas) == p2 - pas && strncmp(lpas, pas, p2 - pas) == 0) {
			    return(FROM_MODERATOR_AUTH);
			}
		    }
		}
	    }

	    if (modEmail) {
		return(FROM_MODERATOR);
	    }
	}
    }
    return(FROM_UNKNOWN);
}

int 
MsgFromCompare(Msg *msg, const char *from, int fromLen)
{
    const char *email;
    int elen;

    email = findHeader(msg, "From:", 0, &elen, "");
    email = parseEMailFromAddress(email, elen, &elen);
    from = parseEMailFromAddress(from, fromLen, &fromLen);
    if (fromLen == elen && strncasecmp(email, from, elen) == 0)
	return(0);
    return(-1);
}

/*
 * PARSEADDRESS() - parse a single address out of the given header
 *
 *	skips any initial commas and whitespace.
 *
 *	return a pointer to the base of the address or NULL, and set
 *	'plen' to the length of the address, with whitespace stripped
 *	and not including any comma.
 */

const char *
parseAddress(const char *field, int len, int *plen)
{
    int ilen;
    int jlen;
    char c1 = 0;

    /*
     * skip separators
     */

    for (ilen = 0; ilen < len; ++ilen) {
	if (field[ilen] != ' ' &&
	    field[ilen] != '\t' &&
	    field[ilen] != '\n' &&
	    field[ilen] != ','
	) {
	    break;
	}
    }

    /*
     * parse address (skip address, handle quoting)
     */

    for (jlen = ilen; jlen < len; ++jlen) {
	if (c1 == 0 && field[jlen] == '<')
	    c1 = '>';
	if (c1 == 0 && field[jlen] == '(')
	    c1 = ')';
	else if (c1 == 0 && field[jlen] == '\"')
	    c1 = '\"';
	else if (c1 && field[jlen] == c1)
	    c1 = 0;

	if ((c1 == 0 && field[jlen] == ',') ||
	    field[jlen] == '\n' ||
	    field[jlen] == '\0'
	) {
	    break;
	}
    }

    /*
     * strip trailing white space, handle end-of-header case
     * (return NULL)
     */

    for (--jlen; jlen >= ilen; --jlen) {
	if (field[jlen] != ' ' && field[jlen] != '\t')
	    break;
    }
    if (ilen == len) {
	*plen = 0;
	return(NULL);
    } else {
	*plen = jlen + 1 - ilen;
	return(field + ilen);
    }
}

/*
 * PARSEEMAILFROMADDRESS() - parse just the email from an address.
 *
 * (comment) email (comment)
 * comment <email> comment
 */

const char *
parseEMailFromAddress(const char *field, int len, int *plen)
{
    int i;

    /*
     * <email>
     */

    for (i = 0; i < len; ++i) {
	if (field[i] == '(') {
	    while (i < len && field[i] != ')')
		++i;
	    continue;
	}
	if (field[i] == '\"') {
	    ++i;
	    while (i < len && field[i] != '"')
		++i;
	    continue;
	}
	if (field[i] == '<') {
	    int j;
	    for (j = ++i; j < len && field[j] != '>'; ++j)
		;
	    if (j < len && field[j] == '>') {
		*plen = j - i;
		return(field + i);
	    }
	}
    }

    /*
     * parse other
     */

    for (i = 0; i < len; ++i) {
	if (field[i] == '(') {
	    while (i < len && field[i] != ')')
		++i;
	    continue;
	}
	if (field[i] == '\"') {
	    ++i;
	    while (i < len && field[i] != '"')
		++i;
	    continue;
	}
	if (field[i] == ' ' || field[i] == '\t' || field[i] == '\n' || field[i] == '\r')
	    continue;
	{
	    int j;
	    for (j = i; j < len; ++j) {
		if (field[j] == ' ' || field[j] == '\t' ||
		    field[j] == '\n' || field[j] == '\r' ||
		    field[j] == ',' || field[j] == '('
		) {
		    break;
		}
	    }
	    *plen = j - i;
	    return(field + i);
	}
    }
    *plen = 0;
    return(NULL);
}

const char *
stripPassword(const char *field, int len, int *plen)
{
    *plen = len;
    return(field);
}

const char *
stripPasswordFromBody(const char *field, int len, int *plen)
{
    const char *ptr = field;

    *plen = len;	/* return whole body on failure */

    /*
     * Skip initial white space
     */

    while (len &&
	(*ptr == ' ' || *ptr == '\t' || *ptr == '\n' || *ptr == '\r')
    ) {
	++ptr;
	--len;
    }

    /*
     * First line of body contains approved ?
     */
    if (len >= 8 && strncasecmp(ptr, "approved", 8) == 0) {
	ptr += 8;
	len -= 8;
	while (len && *ptr != '\n') {
	    --len;
	    ++ptr;
	}

	/*
	 * skip newline at end of approved header,
	 * skip any following headers
	 */
	while (len) {
	    int isHeader = 0;

	    --len;
	    ++ptr;
	    while (len && *ptr != '\n') {
		if (*ptr == ':') {
		    isHeader = 1;
		    break;
		}
		if ((*ptr >= '0' && *ptr <= '9') ||
		    (*ptr >= 'a' && *ptr <= 'z') ||
		    (*ptr >= 'A' && *ptr <= 'Z') ||
		    (*ptr == '-')
		) {
		    ++ptr;
		    --len;
		    continue;
		}
		break;
	    }
	    if (isHeader == 0)
		break;
	    while (len && *ptr != '\n') {
		--len;
		++ptr;
	    }
	}

	/*
	 * Return remainder of body
	 */
	field = ptr;
	*plen = len;
    }
    return(field);
}

/*
 * createQueueFile() - Create and lock a new queue file.
 */

int
createQueueFile(char **pqf, int *pqfo)
{
    char path[256];
    int fd = -1;

    /*
     * create a queue file based on the time of day.  The primary portion
     * of the queue file name is also used for later timeout/removal.  We
     * are limited to 1000 queue files per second.
     */
    {
	struct timeval tv;
	int i;

	for (i = 0; i < 1000; ++i) {
	    gettimeofday(&tv, NULL);
	    snprintf(path, sizeof(path), "%s/spool/x%08lx.%ld",
		BaseDir, 
		(long)tv.tv_sec,
		(tv.tv_usec ? (long)tv.tv_usec / 1000 : (long)i)
	    );
	    *pqfo = strlen(BaseDir) + 7;
	    fd = open(path, O_CREAT|O_EXCL|O_RDWR, 0600);
	    if (fd >= 0) {
		struct stat st;

		st.st_nlink = 1;
		flock(fd, LOCK_EX);
		fstat(fd, &st);
		if (st.st_nlink < 1) {
		    close(fd);
		    fd = -1;
		}
		break;
	    }
	    tv.tv_sec = 0;
	    tv.tv_usec = 1000;
	    select(0, NULL, NULL, NULL, &tv);
	}
    }
    if (fd >= 0) {
	*pqf = dupstr(path);
    }
    return(fd);
}

void 
msgprintf(Msg *msg, const char *ctl, ...)
{
    va_list va;
    char buf[1024];
    int l;

    va_start(va, ctl);
    l = vsnprintf(buf, sizeof(buf), ctl, va);
    va_end(va);
    msg->ms_ReturnBuf = realloc(msg->ms_ReturnBuf, msg->ms_RBLen + l + 1);
    bcopy(buf, msg->ms_ReturnBuf + msg->ms_RBLen, l);
    msg->ms_ReturnBuf[msg->ms_RBLen + l] = 0;
    msg->ms_RBLen += l;
}

