
/*
 * PRUNE.C
 *
 * Remove specified addresses from all or specified lists
 */

#include "defs.h"

#define PHSIZE	1024
#define PHMASK	(PHSIZE-1)

Prototype void RunPrune(const char *listName, int pruneOpt);

void RunPruneFile(const char *listName, int pruneOpt);
long phash(const char *id, int len);
int FindPHash(const char *str, int len, int make);

typedef struct PruneNode {
    struct PruneNode *pn_Next;
    char	*pn_EMail;
    long	pn_Hv;
    int		pn_Len;
} PruneNode;

PruneNode **PruneHash;

void 
RunPrune(const char *listName, int pruneOpt)
{
    DIR *dir;
    struct dirent *den;
    char buf[256];

    if (chdir(BaseDir) < 0)
	fatal(errno, "chdir(%s)", BaseDir);
    if (chdir("lists") < 0)
	fatal(errno, "chdir(%s/spool)", BaseDir);

    PruneHash = calloc(sizeof(PruneNode *), PHSIZE);
    while (fgets(buf, sizeof(buf), stdin) != NULL) {
	int len;
	const char *email = parseEMailFromAddress(buf, strlen(buf), &len);

	(void)FindPHash(email, len, 1);
    }

    if (listName) {
	RunPruneFile(listName, pruneOpt);
    } else {
	if ((dir = opendir(".")) != NULL) {
	    while ((den = readdir(dir)) != NULL) {
		if (den->d_name[0] == '.')
		    continue;
		RunPruneFile(den->d_name, pruneOpt);
	    }
	    closedir(dir);
	}
    }
    /* free hash table and contents */
}

/*
 * Run a queue file.  Note that we must GetMailList() in order
 * to obtain the master lock on the config file to prevent race
 * conditions.
 */

void
RunPruneFile(const char *listName, int pruneOpt)
{
    MailList *list = GetMailList(listName, LOCK_NB);

    printf("list %-25s ", listName);
    fflush(stdout);

    if (list) {
	FILE *fi = OpenListFile(list->ml_ListName, "list", O_CREAT, LOCK_EX|LOCK_NB);

	++list->ml_ListFileLocked;

	if (fi) {
	    char buf[256];

	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		const char *iflags;  
		const char *sem;
		int il;
		int modified = 0;
		char niflags[32];
		char *ptr;

		/*
		 * ignore comments or badly formed entries
		 */

		if (buf[0] == '#')
		    continue;

		if ((iflags = strtok(buf, " \t\r\n")) == NULL)
		    continue;

		if (strlen(iflags) > 30 || iflags[0] == '-')
		    continue;

		strcpy(niflags, iflags);
		for (il = strlen(niflags); il > 0 && niflags[il-1] == '.'; --il)
		    ;
		niflags[il] = 0;

		/*
		 * find counter
		 */
		for (il = 0; niflags[il]; ++il) {
		    if (niflags[il] >= '0' && niflags[il] <= '9')
			break;
		    if (niflags[il] == 'x') {
			niflags[il] = '3';
			modified = 1;
			break;
		    }
		}
		if (niflags[il] == 0) {
		    strcpy(niflags + il, "4");
		}
		ptr = niflags + il;

		if (pruneOpt < 0) {
		    bcopy(ptr + 1, ptr, strlen(ptr));
		    modified = 1;
		}

		if ((sem = strtok(NULL, "\r\n")) == NULL)
		    continue;

		if (pruneOpt > 0) {
		    if (FindPHash(sem, strlen(sem), 0) == 1) {
			if (*ptr > '0') {
			    --*ptr;
			    modified = 1;
			}
		    } else if (pruneOpt == 2) {
			if (*ptr != '0' && *ptr != '4') {
			    ++*ptr;
			    modified = 1;
			}
		    }
		}
		if (modified) {
		    SetListFlags(list, sem, strlen(sem), niflags);
		    printf(" %s(%c)", sem, *ptr);
		    fflush(stdout);
		}
	    }
	    fclose(fi);
	    printf(" ok\n");
	} else {
	    printf("locked, skipping (no list file)\n");
	}
	--list->ml_ListFileLocked;
	CloseMailList(list);
    } else {
	printf("locked, skipping (or not found)\n");
    }
}


long 
phash(const char *id, int len)
{
    long hv = 0xAB5F3D23;

    while (len) {
	hv = (hv << 5) ^ tolower(*id) ^ (hv >> 23);
	++id;
	--len;
    }
    return(hv ^ (hv >> 16));
}

int 
FindPHash(const char *str, int len, int make)
{
    long hv = phash(str, len);
    PruneNode **ppn = &PruneHash[hv & PHMASK];
    PruneNode *pn;
    int r = 0;

    while ((pn = *ppn) != NULL) {
	ppn = &pn->pn_Next;
	if (pn->pn_Len == len && 
	    pn->pn_Hv == hv &&
	    strncasecmp(str, pn->pn_EMail, len) == 0
	) {
	    r = 1;
	    break;
	}
    }
    if (pn == NULL && make) {
	pn = *ppn = malloc(sizeof(PruneNode) + len + 1);
	pn->pn_Next = NULL;
	pn->pn_EMail = (char *)(pn + 1);
	pn->pn_Hv = hv;
	pn->pn_Len = len;
	bcopy(str, pn->pn_EMail, len);
	pn->pn_EMail[len] = 0;
    }
    return(r);
}

