
/*
 * SUBS.C
 */

#include <time.h>

#include "defs.h"

Prototype int stringOk(const char *s, const char **errMsg);
Prototype int stringInString(const char *s, const char *d, int len);
Prototype char *dupstr(const char *str);
Prototype char *dupnstr(const char *str, int len);
Prototype void freestr(char **pstr);
Prototype void fatal(int nerr, const char *ctl, ...);
Prototype void initRandom(int arg);
Prototype int getRandom(int modulo);
Prototype const char *MakeId(Msg *msg);
Prototype int copen(int *pid_p, const char *ctl, ...);
Prototype int copenSendmail(int *pid_p, MailList *list, const char *subSystem, int altQ);
Prototype void cprintf(int fd, const char *ctl, ...);
Prototype void cwrite(int fd, const void *buf, int bytes);
Prototype void cclose(int fd, int pid);
Prototype void *zalloc(int bytes);
Prototype void zfree(void *ptr, int bytes);
Prototype int validEMailAddress(const char *email, int len);

char *
dupstr(const char *str)
{
    char *res = NULL;

    if (str) {
	if ((res = malloc(strlen(str) + 1)) == NULL)
	    fatal(errno, "malloc(%d)", strlen(str) + 1);
	strcpy(res, str);
    }
    return(res);
}

char *
dupnstr(const char *str, int len)
{
    char *res = NULL;

    if (str) {
	if ((res = malloc(len + 1)) == NULL)
	    fatal(errno, "malloc(%d)", len + 1);
	bcopy(str, res, len);
	res[len] = 0;
    }
    return(res);
}

void 
freestr(char **pstr)
{
    if (*pstr != NULL) {
	free(*pstr);
	*pstr = NULL;
    }
}

void
fatal(int nerr, const char *ctl, ...)
{
    va_list va;

    va_start(va, ctl);
    vfprintf(stderr, ctl, va);
    va_end(va);
    if (errno)
	fprintf(stderr, ": %s", strerror(nerr));
    fprintf(stderr, "\n");
    exit(1);
}

void
initRandom(int arg)
{
    double la[3];
    unsigned int x;
    struct timeval tv;

    gettimeofday(&tv, NULL);
    getloadavg(la, 3);
    x = (unsigned int)(la[0] * 1000000.0 + la[1] * 10000.0 + la[2] * 100.0);
    srandom(arg + x + random() + tv.tv_sec + tv.tv_usec);
    random();
    random();
    srandom((unsigned int)time(NULL) + random());
    random();
    random();
}

int
getRandom(int modulo)
{
    int r;

    r = (random() & 0x7FFFFFFF) % modulo;
    return(r);
}

const char *
MakeId(Msg *msg)
{
    struct stat st;
    int l;
    int a = 0;
    static char buf[32];

    bzero(&st, sizeof(st));
    if (msg->ms_Fd >= 0)
	fstat(msg->ms_Fd, &st);
    for (l = 0; l < (sizeof(st) / 4) * 4; ++l)
	a = a ^ ((long *)&st)[l];
    initRandom(a);
    for (l = 0; l < 8; ++l) {
	int c = getRandom(26*2+10);
	if (c < 10)
	    c = c + '0';
	else if (c < 10 + 26)
	    c = c - 10 + 'A';
	else
	    c = c - 36 + 'a';
	buf[l] = c;
    }
    buf[l] = 0;
    return(buf);
}

int
stringOk(const char *s, const char **errMsg)
{
    int r = 0;

    if (s) {
	int l = strlen(s);

	while (*s) {
	    if (*s == '|')
		r = -1;
	    if (*s == ';')
		r = -1;
	    if (*s == '/')
		r = -1;
	    if (*s < 32)
		r = -1;
	    ++s;
	}
	if (l > 128)
	    r = -1;
    }
    return(r);
}

int 
stringInString(const char *s, const char *d, int len)
{
    int l = strlen(s);
    int i;

    for (i = len - l; i >= 0; --i) {
	if (strncasecmp(s, d + i, l) == 0)
	    return(0);
    }
    return(-1);
}

void *
zalloc(int bytes)
{
    void *d = NULL;

    if (bytes) {
	if ((d = malloc(bytes)) == NULL)
	    fatal(errno, "zalloc(%d) failed", bytes);
	bzero(d, bytes);
    }
    return(d);
}

void 
zfree(void *ptr, int bytes)
{
    if (bytes) {
	memset(ptr, -1, bytes);	/* segfault if later ptr indirection */
	free(ptr);
    }
}

/*
 *	SUB-PROCESS ROUTINES
 */

#define FDS_READ	0
#define FDS_WRITE	1

int 
copen(int *pid_p, const char *ctl, ...)
{
    int pid = -1;
    int fds[2];
    int nfd = open("/dev/null", O_RDWR);

    fds[FDS_READ] = -1;
    fds[FDS_WRITE] = -1;

    if (nfd >= 0 && pipe(fds) >= 0) {
	if ((pid = fork()) == 0) {
	    int i;

	    close(fds[FDS_WRITE]);
	    if (fds[FDS_READ] != 0)
		dup2(fds[FDS_READ], 0);
	    if (nfd != 1)
		dup2(nfd, 1);
	    if (nfd != 2)
		dup2(nfd, 2);
	    for (i = 3; i < 20; ++i)
		close(i);
	    execv(ctl, (char * const *)(&ctl) + 1);
	    _exit(1);
	}
    }

    /*
     * close read side and /dev/null desc
     */
    if (fds[FDS_READ] >= 0)
	close(fds[FDS_READ]);

    if (nfd >= 0)
	close(nfd);

    *pid_p = pid;
    return(fds[FDS_WRITE]);
}

int 
copenSendmail(int *pid_p, MailList *list, const char *subSystem, int altQ)
{
    char fopt1[256];
    char fopt2[256];

    if (list)
	snprintf(fopt1, sizeof(fopt1), "-f%s-errors@%s", list->ml_ListName, HostName);
    else
	snprintf(fopt1, sizeof(fopt1), "-fbestserv-errors@%s", HostName);

    if (altQ)
	snprintf(fopt2, sizeof(fopt2), "-od%c", AltQType);
    else
	snprintf(fopt2, sizeof(fopt2), "-odb");
    return(copen(pid_p, SENDMAIL_PATH, "sendmail", "-t", "-oi", fopt1, fopt2, NULL));
}

void 
cprintf(int fd, const char *ctl, ...)
{
    char buf[1024];
    va_list va;
    int l;

    va_start(va, ctl);
    l = vsnprintf(buf, sizeof(buf), ctl, va);
    va_end(va);
    cwrite(fd, buf, l);
}

void 
cwrite(int fd, const void *buf, int bytes)
{
    int r;

    while (fd >= 0 && bytes > 0) {
	r = write(fd, buf, bytes);
	if (r <= 0)
	    break;
	buf = (const char *)buf + r;
	bytes -= r;
    }
}

void 
cclose(int fd, int pid)
{
    if (fd >= 0)
	close(fd);
    if (pid > 0)
	waitpid(pid, NULL, 0);
}

int 
validEMailAddress(const char *email, int len)
{
    int i;
    int at = 0;

    for (i = 0; i < len; ++i) {
	if (email[i] >= 'a' && email[i] <= 'z')
	    continue;
	if (email[i] >= 'A' && email[i] <= 'Z')
	    continue;
	if (email[i] >= '0' && email[i] <= '9')
	    continue;
	switch(email[i]) {
	case '~':
	case '%':
	case '^':
	case '&':
	case '*':
	case '-':
	case '_':
	case '+':
	case '!':
	case '=':
	case ':':
	case '<':
	case '>':
	case '?':
	case '/':
	case ',':
	case '.':
	case '[':
	case ']':
	case '#':
	    continue;
	case '@':
	    at = 1;
	    continue;
	}
	break;
    }
    if (i < len || at == 0)
	return(-1);
    return(0);
}

